<?php

include_once( 'WpAutoPoster_LifeCycle.php' );

class WpAutoPoster_Plugin extends WpAutoPoster_LifeCycle {
	
	protected $logs = null;
	
	/**
	 * Define any tabs to use, if any. Will match formElements using 'tab' field to that tab, but they must be in order.
	 * @return array of tab name data.
	 */
	protected function getTabs( $which = 'main' ) {
		// Use tabs to split up options page. Use formElement 'tab' to match each location.
		if ( $which === 'main' )
			return array(
				array( 'name' => 'Rulesets', 'before' => '<span class="dashicons dashicons-database"></span>' ),
				array( 'name' => 'Logs', 'before' => '<span class="dashicons dashicons-list-view"></span>' ),
				array( 'name' => 'License', 'before' => '<span class="dashicons dashicons-unlock"></span>' ),
				array( 'name' => 'Advanced', 'before' => '<span class="dashicons dashicons-admin-tools"></span>' )
			);
		else if ( $which === 'ruleset' )
			return array(
				array( 'name' => 'General', 'before' => '<span class="dashicons dashicons-admin-settings"></span>' ),
				array( 'name' => 'Post Conditions', 'before' => '<span class="dashicons dashicons-admin-page"></span>' ),
				array( 'name' => 'Post Actions', 'before' => '<span class="dashicons dashicons-edit-page"></span>' ),
				array( 'name' => 'Featured Image', 'before' => '<span class="dashicons dashicons-format-image"></span>' ),
				array( 'name' => 'Post Recycling', 'before' => '<span class="dashicons dashicons-image-rotate"></span>' )
			);
		// Don't use tabs:
		return array();
	}
	
	/**
	 * Modify this function for your plugin settings
	 * @return array of option meta data.
	 */
	public function getOptionMetaData() {

		if ( ! is_admin() ) return null;

		// get some data right away that may be used in several places
		$post_types = get_transient( $this->getPluginTextDomain() . '-post-types' );
		if ( !$post_types ) {
			$post_types = array_diff( get_post_types( array(
				'public' => true 
			), 'names' ), array(
				'attachment' 
			) );
			$post_types = apply_filters( $this->getPluginTextDomain() . '-form-post-types', $post_types );
			set_transient( $this->getPluginTextDomain() . '-post-types', $post_types, MINUTE_IN_SECONDS );
		}

		$post_statuses = get_transient( $this->getPluginTextDomain() . '-post-statuses' );
		if ( !$post_statuses ) {
			$post_statuses = array_diff( get_post_stati(), array(
				'auto-draft',
				'inherit',
				'request-pending',
				'request-confirmed',
				'request-failed',
				'request-completed'
			) );
			$post_statuses = apply_filters( $this->getPluginTextDomain() . '-form-post-statuses', $post_statuses );
			set_transient( $this->getPluginTextDomain() . '-post-statuses', $post_statuses, MINUTE_IN_SECONDS );
		}

		$maxEntries = $this->getOption( 'SelectMaxEntries', 1000 );
		
		$authors = get_transient( $this->getPluginTextDomain() . '-post-authors' );
		if ( !$authors ) {
			$args       = apply_filters( $this->getPluginTextDomain() . '-form-post-authors', array(
				'capability' => array( 'edit_posts' ),
				'fields' => array(
					'ID',
					'display_name' 
				),
				'number' => $maxEntries 
			) );

			// Capability queries were only introduced in WP 5.9.
			if ( version_compare( $GLOBALS['wp_version'], '5.9-alpha', '<' ) ) {
    				$args['who'] = 'authors';
    				unset( $args['capability'] );
			}

			$user_query = new \WP_User_Query( $args );
			$authors    = $user_query->get_results();
			$authors = apply_filters( $this->getPluginTextDomain() . '-form-post-authors2', $authors );
			set_transient( $this->getPluginTextDomain() . '-post-authors', $authors, MINUTE_IN_SECONDS );
		}

		// Set options into array
		
		$metaData = array(
			'ID' => array(
				'description' => esc_html__( 'ID:', $this->getPluginTextDomain() ),
				'formElement' => array(
					'type' => 'hidden',
					'value' => '',
					'tab' => 'Rulesets',
					'useIndex' => 'Rulesets',
					'indexRequired' => true 
				) 
			),
			// GENERAL
			'Status' => array(
				'description' => esc_html__( 'Status:', $this->getPluginTextDomain() ),
				'formElement' => array(
					'type' => 'text',
					'value' => '',
					'default' => 'disabled',
					'tab' => 'General',
					'useIndex' => 'Rulesets',
					'params' => 'readonly class="wp-auto-poster-input-transparent"' 
				) 
			),
			'Name' => array(
				'description' => esc_html__( 'Name:', $this->getPluginTextDomain() ),
				'formElement' => array(
					'type' => 'text',
					'value' => '',
					'tab' => 'General',
					'useIndex' => 'Rulesets',
					'subtext' => 'An identifier for this ruleset, if desired.',
					'tooltip' => '<h3>' . esc_html__( 'Name', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'A moniker to summarize this ruleset.', $this->getPluginTextDomain() ) . '</li></ul>' 
				) 
			),
			'FrequencyHR' => array(
				'formElement' => array(
					'type' => 'hr',
					'tab' => 'General',
					'useIndex' => 'Rulesets',
				)
			),
			'Frequency' => array(
				'description' => esc_html__( 'Event Frequency:', $this->getPluginTextDomain() ),
				'formElement' => array(
					'type' => 'text',
					'value' => '',
					'placeholder' => 'required',
					'tab' => 'General',
					'useIndex' => 'Rulesets',
					'subtext' => esc_html__( 'How often to check this Ruleset, in format DdHhMmSs, such as 1d, 14h, 30m, 20s or even 1d8h10m30s.', $this->getPluginTextDomain() ),
					'indexRequired' => true,
					'tooltip' => '<h3>' . esc_html__( 'Event Frequency', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This setting is used to schedule an event in WordPress. When this ruleset event is triggered by a cron check, WP Auto Poster will check the ruleset settings to see if any posts need to be acted upon.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'By default, WordPress typically checks for scheduled events whenever a page is loaded.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Standard format is DdHhMmSs, such as 1d, 14h, 30m, 20s or even 1d8h10m30s', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Time range format is DdHhMmSs-DdHhMmSs, such as 10h-20h, which will randomly choose the next event time between 10 and 20 hours.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Delay format is DdHhMmSs+DdHhMmSs, such as 5h+10h, which will delay the first event by 5 hours then schedule further events every 10 hours.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Altogether example: 5h30m+10h-20h will delay the first event by 5 hours 30 minutes, then each further event will be randomly set between 10 and 20 hours from the previous event.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'You may also specify multiple possible Event Frequencies, such as "10m,30m,4h" and one will be chosen at random each time for the next event.', $this->getPluginTextDomain() ) . '</li></ul>' 
				) 
			),
			'RestartOnPublish' => array(
				'description' => esc_html__( 'Restart Frequency on Publish?', $this->getPluginTextDomain() ),
				'formElement' => array(
					'type' => 'select',
					'subtext' => esc_html__( 'Next frequency check will be reset when a selected post type is published manually or by any other source.', $this->getPluginTextDomain() ),
					'value' => $post_types,
					'default' => '',
					'tab' => 'General',
					'useIndex' => 'Rulesets',
					'params' => 'multiple size="3" class="wapSelect"',
					'tooltip' => '<h3>' . esc_html__( 'Restart Frequency on Publish', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Example: The Event Frequency is every 24 hours, but 20 hours since the last ruleset event triggered you decide to publish a post manually. The next ruleset event is scheduled to trigger in 4 hours.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'If this setting is NOT checked, the next ruleset event will still occur in 4 hours, as scheduled.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'If this setting IS checked, the next ruleset event will restart and be set to 24 hours from when you manually published the post.', $this->getPluginTextDomain() ) . '</li></ul>' 
				) 
			),
			'Inclusions' => array(
				'description' => esc_html__( 'Inclusion Dates:', $this->getPluginTextDomain() ),
				'formElement' => array(
					'type' => 'text',
					'value' => '',
					'tab' => 'General',
					'useIndex' => 'Rulesets',
					'subtext' => esc_html__( 'Only parse this ruleset on these specific dates and/or times. Separate multiple entries with commas.', $this->getPluginTextDomain() ),
					'tooltip' => '<h3>' . esc_html__( 'Inclusion Dates', $this->getPluginTextDomain() ) . '</h3>' . esc_html__( 'Recognizes the following date and time formats:', $this->getPluginTextDomain() ) . '<ul><li>' . esc_html__( 'YYYY-MM-DD Example: 2020-12-25 is the 25th of December, 2020.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'YYYY-MM Example: 2020-12 is the entire month of December, 2020.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'MM-DD Example: 12-25 is the 25th of December of every year.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'YYYY Example: 2020 is the entire year of 2020.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'DD Example: 25 is the 25th of every month.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'M (Three letter month) Example: Dec is the entire month of December. You can use slash for more than one month, such as Jan/Feb/Mar.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'D (Three letter day) Example: Mon is every Monday. You can use slash for more than one day, such as Mon/Wed/Fri.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'hhhh-hhhh is a 24 hour time range. Example: 0900-1700 is between the hours of 9am and 5pm. You may attach the time range after any date format. Examples:', $this->getPluginTextDomain() ) . '<ul><li>' . esc_html__( 'Nov/Dec:0900-1700 is between the hours of 9am and 5pm only in November and December.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Mon/Wed/Fri:0900-1700 is between the hours of 9am and 5pm only on Mondays, Wednesdays and Fridays.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( '12-25:0900-1700 is between the hours of 9am and 5pm only on Dec 25th.', $this->getPluginTextDomain() ) . '</li></ul></li><li>' . esc_html__( 'Relative formats for YYYY-MM-DD dates. Examples: \'last day of this month\' and \'second saturday of July\'', $this->getPluginTextDomain() ) . '</ul>',
					'fullWidth' => true 
				) 
			),
			'Exclusions' => array(
				'description' => esc_html__( 'Exclusion Dates:', $this->getPluginTextDomain() ),
				'formElement' => array(
					'type' => 'text',
					'value' => '',
					'tab' => 'General',
					'useIndex' => 'Rulesets',
					'subtext' => esc_html__( 'Any specific dates and/or times to NOT parse this ruleset. Separate multiple entries with commas.', $this->getPluginTextDomain() ),
					'tooltip' => '<h3>' . esc_html__( 'Exclusion Dates', $this->getPluginTextDomain() ) . '</h3>' . esc_html__( 'Recognizes the following date and time formats:', $this->getPluginTextDomain() ) . '<ul><li>' . esc_html__( 'YYYY-MM-DD Example: 2020-12-25 is the 25th of December, 2020.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'YYYY-MM Example: 2020-12 is the entire month of December, 2020.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'MM-DD Example: 12-25 is the 25th of December of every year.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'YYYY Example: 2020 is the entire year of 2020.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'DD Example: 25 is the 25th of every month.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'M (Three letter month) Example: Dec is the entire month of December. You can use slash for more than one month, such as Jan/Feb/Mar.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'D (Three letter day) Example: Mon is every Monday. You can use slash for more than one day, such as Mon/Wed/Fri.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'hhhh-hhhh is a 24 hour time range. Example: 0900-1700 is between the hours of 9am and 5pm. You may attach the time range after any date format. Examples:', $this->getPluginTextDomain() ) . '<ul><li>' . esc_html__( 'Nov/Dec:0900-1700 is between the hours of 9am and 5pm only in November and December.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Mon/Wed/Fri:0900-1700 is between the hours of 9am and 5pm only on Mondays, Wednesdays and Fridays.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( '12-25:0900-1700 is between the hours of 9am and 5pm only on Dec 25th.', $this->getPluginTextDomain() ) . '</li></ul></li><li>' . esc_html__( 'Relative formats for YYYY-MM-DD dates. Examples: \'last day of this month\' and \'second saturday of July\'', $this->getPluginTextDomain() ) . '</ul>',
					'fullWidth' => true 
				) 
			)
		);

		$metaData[ 'MaxDailyPosts' ] = array(
			'description' => esc_html__( 'Max Posts per Day:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '0',
				'params' => 'max=10000 min=0 step=1',
				'tab' => 'General',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Max number of posts WAP will act upon each day from this ruleset. 0 for no limit.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Max Posts per Day', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Max number of posts WAP will act upon each day from this ruleset. You can enter 0 for no limit.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'TriggerOnTransitionHR' ] = array(
			'formElement' => array(
				'type' => 'hr',
				'tab' => 'General',
				'useIndex' => 'Rulesets',
			) 
		);

		$metaData[ 'TriggerOnTransition' ] = array(
			'description' => esc_html__( 'Trigger Ruleset on publish?', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'The Ruleset can also be automatically triggered when a post is published outside of WAP.', $this->getPluginTextDomain() ),
				'value' => '',
				'default' => 'off',
				'tab' => 'General',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Trigger Ruleset on Publish?', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Can also trigger the Ruleset when a post is published outside of WAP that matches the Post Conditions. Ruleset must be enabled.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		// POST CONDITIONS
		$metaData[ 'MaxNumberPosts' ] = array(
			'description' => esc_html__( 'Max Posts per Query:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '1',
				'params' => 'max=10000 min=1 step=1',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Max number of posts to query during each ruleset check.', $this->getPluginTextDomain() ),
				'indexRequired' => true,
				'tooltip' => '<h3>' . esc_html__( 'Max Posts per Query', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Max number of posts to query for during each ruleset check. Be aware that, depending on your server, a large query amount may run out of memory or reach any time constraints set by your server configuration.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'IgnoreStickies' ] = array(
			'description' => esc_html__( 'Ignore Sticky Posts?', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Queries will ignore sticky posts.', $this->getPluginTextDomain() ),
				'value' => '',
				'default' => 'on',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Ignore Sticky Posts', $this->getPluginTextDomain() ) . '</h3><ul><li>' . sprintf( '<a target="_blank" href="%s">', esc_url( 'https://codex.wordpress.org/Sticky_Posts' ) ) . esc_html__( 'Sticky posts', $this->getPluginTextDomain() ) . '</a>' . esc_html__( ' are a WordPress feature for standard posts, and are automatically placed at the top of the front page of posts.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'SuppressFilters' ] = array(
			'description' => esc_html__( 'Suppress Filters?', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Queries will suppress filters to prevent other plugins from modifying the results.', $this->getPluginTextDomain() ),
				'value' => '',
				'default' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Suppress Filters', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Queries will suppress filters to prevent other plugins from modifying the results, such as WPML which would modify the query to only return posts that match the current language. Usually this is not required.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'LimitPostTypes' ] = array(
			'description' => esc_html__( 'Limit check to Post Types:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Queries will be limited to selected post types.', $this->getPluginTextDomain() ),
				'value' => $post_types,
				'default' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'params' => 'multiple size="5" class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Limit check to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'These will limit the eligible post query to the post types selected. If nothing is selected, the WordPress default \'post\' will be used.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$ps = array( '(recycle only)' => '(recycle only)' ) + $post_statuses;
		$metaData[ 'LimitPostStatuses' ] = array(
			'description' => esc_html__( 'Limit check to Post Statuses:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Queries will be limited to selected post statuses.', $this->getPluginTextDomain() ),
				'value' => $ps,
				'default' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'params' => 'multiple size="5" class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Limit check to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'These will limit the eligible post query to the post statuses selected. If (recycle only) or nothing is selected, WAP will skip direct to Post Recycling.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'LimitPostAuthors' ] = array(
			'description' => esc_html__( 'Limit check to Authors:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Queries will be limited to selected authors.', $this->getPluginTextDomain() ),
				'value' => $authors,
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'params' => 'multiple size="5" class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Limit check to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'These will limit the eligible post query to the authors selected.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$support       = $this->getOption( 'Support' );
		$useTaxonomies = $support ? $this->getOption( 'UseTaxonomies', array ()) : array(
			'category',
			'post_tag' 
		);
		
		foreach ( $useTaxonomies as $useTaxonomy ) {
			if ( false === taxonomy_exists( $useTaxonomy ) )
				continue;
			$taxonomy_entries = get_transient( $this->getPluginTextDomain() . '-' . $useTaxonomy );
			
			if ( !$taxonomy_entries ) {
				
				if ( wp_count_terms( $useTaxonomy, array(
					'hide_empty' => 0 
				) ) > $maxEntries )
					$args = array(
						'taxonomy' => $useTaxonomy,
						'hide_empty' => 0,
						'hierarchical' => false,
						'number' => $maxEntries 
					);
				else
					$args = array(
						'taxonomy' => $useTaxonomy,
						'hide_empty' => 0 
					);
				
				$args = apply_filters( $this->getPluginTextDomain() . '-form-taxonomy', $args );
				
				// defaults to orderby name ASC (requires WP 4.5+)
				$taxonomy_entries = get_terms( $args );
				if ( is_wp_error( $taxonomy_entries ) ) {
					$msg = sprintf( esc_html__( 'WP error in WAP %s - %s - get_terms args were: %s', $this->getPluginTextDomain() ), __FUNCTION__, $taxonomy_entries->get_error_message(), print_r( $args, true ) );
					echo "<div class='error'><p>" . $msg . "</p></div>";
				} else {
					// set up hierarchy of terms
					$taxonomy_hierarch = array();
					$this->sortTermsHierarchicaly( $taxonomy_entries, $taxonomy_hierarch );
					$taxonomy_entries = array();
					// convert term objects to id->name array
					$this->convertTerms( $taxonomy_hierarch, $taxonomy_entries );
					set_transient( $this->getPluginTextDomain() . '-' . $useTaxonomy, $taxonomy_entries, MINUTE_IN_SECONDS );
				}
			}

			if ( !in_array( $useTaxonomy, array(
				'category',
				'post_tag' 
			) ) )
					$limitValue = array(
						'IN' => 'In',
						'NOT IN' => 'Not In',
						'EXISTS' => 'Exists',
						'NOT EXISTS' => 'Not Exists' 
					);
			else
					$limitValue = array(
						'IN' => 'In',
						'NOT IN' => 'Not In'
					);

			$metaData[ 'Limit-' . $useTaxonomy ] = array(
				'description' => sprintf( esc_html__( '%s limit operator:', $this->getPluginTextDomain() ), $useTaxonomy ),
				'combine' => $useTaxonomy,
				'formElement' => array(
					'type' => 'select',
					'subtext' => sprintf( esc_html__( 'Which operator to use on the %s taxonomy using any selected entries below.', $this->getPluginTextDomain() ), $useTaxonomy ),
					'value' => $limitValue,
					'default' => 'IN',
					'tab' => 'Post Conditions',
					'useIndex' => 'Rulesets',
					'params' => 'class="wapSelect wapSelectBlock"',
					'tooltip' => '<h3>' . sprintf( esc_html__( '%s action', $this->getPluginTextDomain() ), $useTaxonomy ) . '</h3><ul><li>' . sprintf( esc_html__( 'Limit the post search for the %s taxonomy using this operator for any selected terms below.', $this->getPluginTextDomain() ), $useTaxonomy ) . '</li></ul>' 
				) 
			);
			
			$metaData[ $useTaxonomy ] = array(
				'description' => sprintf( esc_html__( 'Limit check to %s:', $this->getPluginTextDomain() ), $useTaxonomy ),
				'formElement' => array(
					'type' => 'select',
					'subtext' => sprintf( esc_html__( 'Queries will be limited to selected %s entries.', $this->getPluginTextDomain() ), $useTaxonomy ),
					'value' => $taxonomy_entries,
					'default' => '',
					'tab' => 'Post Conditions',
					'useIndex' => 'Rulesets',
					'params' => 'multiple size="5" class="wapSelect"',
					'tooltip' => '<h3>' . esc_html__( 'Limit check to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'These will limit the eligible post query to the taxonomy terms selected. Set the taxonomies to display in the Advanced tab.', $this->getPluginTextDomain() ) . '</li></ul>' 
				) 
			);
			if ( !in_array( $useTaxonomy, array(
				'category',
				'post_tag' 
			) ) )
				$metaData[ $useTaxonomy ][ 'formElement' ][ 'tooltip' ] .= '<ul><li>' . esc_html__( 'Note that if custom taxonomies are selected, WordPress will not limit the Post Types.', $this->getPluginTextDomain() ) . '</li></ul>';
		}
		
		$metaData[ 'LimitPostSearch' ] = array(
			'description' => esc_html__( 'Limit check to Keyword Search:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Queries will be limited using these search terms.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Limit check to Keyword Search', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This invokes the ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">', esc_url( 'https://developer.wordpress.org/reference/classes/wp_query/#search-parameters' ) ) . esc_html__( 'WordPress search parameter', $this->getPluginTextDomain() ) . '</a>. ' . esc_html__( 'A couple of notes about their search:', $this->getPluginTextDomain() ) . '<br><br><ul><li>' . esc_html__( 'WordPress searches the post title, post content, and post excerpt.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Prepending a term with a hyphen will exclude posts matching that term.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Multiple words are searched using the AND parameter by default.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Note there is a WordPress ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">%s</a> ', esc_url( 'https://core.trac.wordpress.org/ticket/43634' ), esc_html__( 'programming bug', $this->getPluginTextDomain() ) ) . esc_html__( 'if using all negative search terms and the number of negative search terms is greater than 9.', $this->getPluginTextDomain() ) . '</ul></li></ul>',
				'fullWidth' => true 
			) 
		);

		$metaData[ 'CustomSearch' ] = array(
			'description' => esc_html__( 'Limit Keyword Search to:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'value' => array(
					'title'		=> esc_html__( 'Post Title', $this->getPluginTextDomain() ),
					'content'	=> esc_html__( 'Post Content', $this->getPluginTextDomain() ),
					'excerpt'	=> esc_html__( 'Post Excerpt', $this->getPluginTextDomain() ),
				),
				'default' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Limit Keyword Search to', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This will modify the WordPress search to target only the specified content fields. By default WordPress searches them all.', $this->getPluginTextDomain() ) . '</li></ul>',
				'params' => 'multiple',
				'fullSpace' => true 
			) 
		);

		$metaData[ 'LimitPostIDs' ] = array(
			'description' => esc_html__( 'Limit check by Post IDs:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Queries will be limited using these post IDs.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Limit check by Post IDs', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'You can limit the query to specific post IDs if needed. You may instead use negative post ID numbers to exclude them from the query. Do not use both as they are mutually exclusive.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Example 1: 4,12,17 will limit query to post IDs 4, 12 and 17.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Example 2: -4,-12,-17 will limit query to exclude post IDs 4, 12 and 17.', $this->getPluginTextDomain() ) . '</li></ul>',
				'fullWidth' => true 
			) 
		);


		$metaData[ 'LimitFeaturedImages' ]    = array(
			'description' => esc_html__( 'Limit by Featured Images:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Queries will be limited based on post thumbnails, aka featured images.', $this->getPluginTextDomain() ),
				'value' => array(
					'No'	=> esc_html__( 'No', $this->getPluginTextDomain() ),
					'Has'	=> esc_html__( 'Has featured image', $this->getPluginTextDomain() ),
					'Missing'	=> esc_html__( 'Missing featured image', $this->getPluginTextDomain() ),
				),
				'default' => 'No',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Limit check to Featured Images', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'No - No limiting by thumbnails.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Has featured image - Limit check to posts that have featured images.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Missing featured image - Limit check to posts without featured images.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'MinPostAge' ]      = array(
			'description' => esc_html__( 'Minimum Post Age:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '0',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Posts must be older than this number of days to be eligible. Enter 0 to ignore.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Minimum Post Age', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Posts must be older than this number of days to be eligible. Enter 0 or leave blank to ignore. Also accepts any ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">%s</a> ', esc_url( 'https://www.php.net/manual/en/function.strtotime.php'), 'strtotime' ) . esc_html__( 'date format', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'MaxPostAge' ]      = array(
			'description' => esc_html__( 'Maximum Post Age:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '0',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Posts must be newer than this number of days to be eligible. Enter 0 or leave blank to ignore.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Maximum Post Age', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Posts must be newer than this number of days to be eligible. Enter 0 or leave blank to ignore. Also accepts any ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">%s</a> ', esc_url( 'https://www.php.net/manual/en/function.strtotime.php'), 'strtotime' ) . esc_html__( 'date format', $this->getPluginTextDomain() ) . '</li></ul>' 

			) 
		);

		$metaData[ 'CommentsOperator' ] = array(
			'description' => esc_html__( 'Required Comments:', $this->getPluginTextDomain() ),
			'combine' => 'RequiredComments',
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'The operator to use for the Required Comments above.', $this->getPluginTextDomain() ),
				'value' => array(
					'='	=> esc_html__( '=', $this->getPluginTextDomain() ),
					'!='	=> esc_html__( '!=', $this->getPluginTextDomain() ),
					'gt='	=> esc_html__( '>=', $this->getPluginTextDomain() ),
					'lt='	=> esc_html__( '<=', $this->getPluginTextDomain() ),
				),
				'default' => '&gt;=',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelectSmallInline"',
			) 
		);
		$metaData[ 'RequiredComments' ]  = array(
			'formElement' => array(
				'type' => 'number',
				'value' => '0',
				'params' => 'max=36500 min=0 step=1',
				'subtext' => esc_html__( '.', $this->getPluginTextDomain() ),
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'The number of post comments to be eligible.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Required Comments', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Enter a number if you wish to limit the eligible posts results to the number of post comments.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$maxMetaFields = apply_filters( $this->getPluginTextDomain() . '-metafield-cond', 1, $ruleset );
		for ( $x = 1; $x <= $maxMetaFields; $x++ ) {

		$metaData[ 'MetaFieldCondName' . $x ]      = array(
			'combine' => 'MetaFieldCompare',
			'description' => esc_html__( 'Limit by Custom Meta Field:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'placeholder' => 'name',
				'params' => 'class="wapSelectMediumInline"',
				'subtext' => esc_html__( 'Custom meta field to check.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Custom Meta Field', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Can compare a meta field name to a certain value. Ex: NAME != Bob will return any post with a custom meta field NAME that does not have the value \'Bob\'', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'If name is left blank, any name can compare to the value. Ex: [blank] = Steve will return any post with any custom meta field that has the value \'Steve\'', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'If value is left blank, any value for the name will match, regardless of comparison. Ex: NAME will return any post with a custom meta field called \'NAME\'', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'MetaFieldCondCompare' . $x ] = array(
			'combine' => 'MetaFieldValue',
			'formElement' => array(
				'type' => 'select',
				'value' => array(
					'='	=> esc_html__( '=', $this->getPluginTextDomain() ),
					'!='	=> esc_html__( '!=', $this->getPluginTextDomain() ),
					'gt'	=> esc_html__( '>', $this->getPluginTextDomain() ),
					'gt='	=> esc_html__( '>=', $this->getPluginTextDomain() ),
					'lt'	=> esc_html__( '<', $this->getPluginTextDomain() ),
					'lt='	=> esc_html__( '<=', $this->getPluginTextDomain() ),
					'LIKE'	=> esc_html__( 'LIKE', $this->getPluginTextDomain() ),
					'NOT LIKE'	=> esc_html__( 'NOT LIKE', $this->getPluginTextDomain() ),
				),
				'default' => '=',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelectSmallInline"',
			) 
		);
		$metaData[ 'MetaFieldCondValue' . $x ]      = array(
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'placeholder' => 'value',
				'params' => 'class="wapSelectMediumInline"',
			) 
		);
		}	// end for x


		$qOrder                        = array(
			'CommentCount' => esc_html__( 'Comment Count', $this->getPluginTextDomain() ),
			'Newest' => esc_html__( 'Newest Posts', $this->getPluginTextDomain() ),
			'Oldest' => esc_html__( 'Oldest Posts', $this->getPluginTextDomain() ),
			'Random' => esc_html__( 'Random Posts', $this->getPluginTextDomain() ),
			'SearchRelevance' => esc_html__( 'Search Relevance', $this->getPluginTextDomain() ) 
		);
		$metaData[ 'QueryOrderBy' ]    = array(
			'description' => esc_html__( 'Order of Query Results:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'How eligible posts returned by the query will be ordered.', $this->getPluginTextDomain() ),
				'value' => $qOrder,
				'default' => 'Oldest',
				'tab' => 'Post Conditions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Order of Query Results', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Comment Count - Post results will be ordered from most commented on to lowest.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Newest Posts - Post results will be ordered from newest to oldest.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Oldest Posts - Post results will be ordered from oldest to newest.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Random Posts - Post results will be randomly ordered.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Search Relevance - Post results will be ordered by relevance to the Keyword Search.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		/* --- POST ACTIONS --- */

		$metaData[ 'ClonePost' ] = array(
			'description' => esc_html__( 'Clone Post:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Will clone the eligible posts then apply Post Actions to the clones.', $this->getPluginTextDomain() ),
				'value' => '',
				'default' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Clone Post', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Will clone the eligible posts then apply Post Actions to the clones. This is useful to create new posts from draft/pending post template(s). Cloned posts default to draft status so be sure to set a new post status if desired.', $this->getPluginTextDomain() ) . '</li></ul>',
			) 
		);
		
		$metaData[ 'SetPostSticky' ] = array(
			'description' => esc_html__( 'Set Post Sticky to:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Set the post sticky of eligible posts. This only works on standard WordPress \'post\' types.', $this->getPluginTextDomain() ),
				'value' => array(
					 0 => 'inherit',
					'Stick' => 'Stick',
					'Unstick' => 'Unstick' 
				),
				'default' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Set Post Sticky to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Set post sticky of eligible posts when triggered. Inherit means no change.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		array_unshift( $post_types, 'inherit' );
		$metaData[ 'SetPostType' ] = array(
			'description' => esc_html__( 'Set Post Type to:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Set the post type of eligible posts.', $this->getPluginTextDomain() ),
				'value' => $post_types,
				'default' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Set Post Type to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Set post type of eligible posts when triggered. Inherit means no change.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		// add and push these values to the top
		$post_statuses = array( 'inherit' => 'inherit', 'trash' => 'trash' ) + $post_statuses;
		$metaData[ 'SetPostStatus' ] = array(
			'description' => esc_html__( 'Set Post Status to:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Set the post status of eligible posts.', $this->getPluginTextDomain() ),
				'value' => $post_statuses,
				'default' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Set Post Status to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Set post status of eligible posts when triggered. Inherit means no change.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'If setting a draft or pending post to \'publish\', WordPress will also update the publish date of the post unless you choose to keep the original date.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'If you are publishing a previously published post, WordPress will not update the post published date. Use Post Recycling instead if you want the published date updated.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'SetPostDate' ] = array(
			'description' => esc_html__( 'Set Post Publish Date to:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Set the post publish date of eligible posts.', $this->getPluginTextDomain() ),
				'value' => array(
					 0 => 'inherit',
					'Keep Time' => 'Keep Time',
					'Current Time' => 'Current Time',
				),
				'default' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Set Post Publish Date to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Set post date of eligible posts when triggered. Inherit means using the default WordPress action, if any. Typically, this would update the post publish date when a NEW draft or pending post is published and the post_date_gmt has not already been set.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Keep Time will ensure WordPress does not change the post publish date.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Current Time will ensure WordPress changes the post publish date to the current time when triggered.', $this->getPluginTextDomain() ) . '</li></ul><p>' . esc_html__( 'Note that updating a post will always update its last_modified time, even if the published date does not change.', $this->getPluginTextDomain() ) . '</p>'
			) 
		);
		
		array_unshift( $authors, 'inherit' );
		$authors[99999] = '(random)';
		$metaData[ 'SetPostAuthor' ] = array(
			'description' => esc_html__( 'Set Author to:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Set the author of eligible posts.', $this->getPluginTextDomain() ),
				'value' => $authors,
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Set Author to...', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Set author of eligible posts when triggered. Inherit means no change.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'FindReplace' ] = array(
			'description' => esc_html__( 'Find and Replace:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'textarea',
				'value' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'placeholder' => htmlentities( 'find => replace' ),
				'tooltip' => '<h3>' . esc_html__( 'Find and Replace', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Will search and replace in post title, content, and excerpt unless limited.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'One rule per line in X => Y format. Case is preserved. Supports spintax for replacement.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Spintax is resolved once and replaces every find.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Note if the title changes so does the post slug, so the post permalink may change.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Example 1, replace every word cat with dog: "cat => dog"', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Example 2, remove every word cat: "cat =>"', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Example 3, replace every word cat with random spintax: "cat => {dog|elephant|weasel}"', $this->getPluginTextDomain() ) . '</li></ul>Special Keywords:<ul><li>' . esc_html__( 'Remove every sentence containing the word cat: "cat => REMOVE_SENTENCE"', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Replace every %%date%% with the current date: "%%date%% => DATE_FORMAT"', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Replace every %%time%% with the current time: "%%time%% => TIME_FORMAT"', $this->getPluginTextDomain() ) . '</li></ul>',
				'fullWidth' => true 
			) 
		);

		$metaData[ 'CustomFindReplace' ] = array(
			'description' => esc_html__( 'Limit Find and Replace to:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'value' => array(
					'title'		=> esc_html__( 'Post Title', $this->getPluginTextDomain() ),
					'content'	=> esc_html__( 'Post Content', $this->getPluginTextDomain() ),
					'excerpt'	=> esc_html__( 'Post Excerpt', $this->getPluginTextDomain() ),
				),
				'default' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Limit Find and Replace to', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This will modify the find and replace to target only the specified content fields.', $this->getPluginTextDomain() ) . '</li></ul>',
				'params' => 'multiple',
				'fullSpace' => true 
			) 
		);

		$metaData[ 'PostActionNotify' ] = array(
			'description' => esc_html__( 'Notify Post Author:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Will send the post author an email with any changes made.', $this->getPluginTextDomain() ),
				'value' => '',
				'default' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Notify Post Author', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This will send an email to the post author from the site admin notifying them of any changes made.', $this->getPluginTextDomain() ) . '</li></ul>',
			) 
		);

		$metaData[ 'PostActionsTaxonomyHR' ] = array(
			'formElement' => array(
				'type' => 'hr',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Note: Changing any taxonomies below does not update the post modified date.', $this->getPluginTextDomain() )
			) 
		);


		foreach ( $useTaxonomies as $useTaxonomy ) {
			if ( false === taxonomy_exists( $useTaxonomy ) )
				continue;
			
			$metaData[ 'Action-' . $useTaxonomy ] = array(
				'description' => sprintf( esc_html__( '%s action:', $this->getPluginTextDomain() ), $useTaxonomy ),
				'combine' => $useTaxonomy,
				'formElement' => array(
					'type' => 'select',
					'subtext' => sprintf( esc_html__( 'How to apply any selected %s changes below.', $this->getPluginTextDomain() ), $useTaxonomy ),
					'value' => array(
						'None' => 'None',
						'Add' => 'Add',
						'Delete' => 'Delete',
						'Replace' => 'Replace' 
					),
					'default' => '',
					'tab' => 'Post Actions',
					'useIndex' => 'Rulesets',
					'params' => 'class="wapSelect wapSelectBlock"',
					'tooltip' => '<h3>' . sprintf( esc_html__( '%s action', $this->getPluginTextDomain() ), $useTaxonomy ) . '</h3><ul><li>' . sprintf( esc_html__( 'Whether to add, delete, or replace selected %s taxonomy terms for eligible posts.', $this->getPluginTextDomain() ), $useTaxonomy ) . '</li></ul>' 
				) 
			);
			
			
			
			$taxonomy_entries                     = get_transient( $this->getPluginTextDomain() . '-' . $useTaxonomy );
			$metaData[ 'Change-' . $useTaxonomy ] = array(
				'description' => sprintf( esc_html__( '%s changes:', $this->getPluginTextDomain() ), $useTaxonomy ),
				'formElement' => array(
					'type' => 'select',
					'subtext' => sprintf( esc_html__( 'The %s terms to act upon using the action above.', $this->getPluginTextDomain() ), $useTaxonomy ),
					'value' => $taxonomy_entries,
					'default' => '',
					'tab' => 'Post Actions',
					'useIndex' => 'Rulesets',
					'params' => 'multiple size="5" class="wapSelect"',
					'tooltip' => '<h3>' . esc_html__( 'Change this taxonomy', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Change taxononmy terms of eligible posts when triggered. No selection means no change.', $this->getPluginTextDomain() ) . '</li></ul>' 
				) 
			);
		}

		$metaData[ 'PostActionsTaxonomyHR2' ] = array(
			'formElement' => array(
				'type' => 'hr',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
			) 
		);

		$maxMetaFields = apply_filters( $this->getPluginTextDomain() . '-metafield-actions', 1, $ruleset );
		for ( $x = 1; $x <= $maxMetaFields; $x++ ) {

		$metaData[ 'MetaFieldActionName' . $x ]      = array(
			'combine' => 'MetaFieldCompare',
			'description' => esc_html__( 'Set Custom Meta Field:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'placeholder' => 'name',
				'params' => 'class="wapSelectMediumInline"',
				'subtext' => esc_html__( 'Save a custom meta key and value.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Custom Meta Field', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Set a value for a custom meta field.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'MetaFieldActionValue' . $x ]      = array(
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'Post Actions',
				'useIndex' => 'Rulesets',
				'placeholder' => 'value',
				'params' => 'class="wapSelectMediumInline"',
			) 
		);
		}


		/* --- FEATURED IMAGE --- */

		$metaData[ 'ModifyFeaturedImage' ] = array(
			'description' => esc_html__( 'Modify Featured Image:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Whether to Modify the Existing Featured Image.', $this->getPluginTextDomain() ),
				'value' => array(
					'No' => 'No',
					'Rename' => 'Rename',
					'Remove' => 'Remove',
					'Delete' => 'Delete'
				),
				'default' => 'No',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Modify Featured Image', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Rename will change the Featured Image file name to match the post title.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Remove will remove the Featured Image from the post but the image file will remain.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Delete will remove the Featured Image from the post and delete the image file completely.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'FindFeaturedImage' ] = array(
			'description' => esc_html__( 'Find Featured Image:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Check post content for any image to make Featured.', $this->getPluginTextDomain() ),
				'value' => array(
					'No' => 'No',
					'Yes' => 'Yes',
					'Only if missing' => 'Only if missing',
				),
				'default' => 'No',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Find Featured Image', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Look for any images in the post content and import the first one found that is bigger than the declared thumbnail size in Settings -> Media. Imported image filenames will be based on the post title automatically.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'FeaturedImageHR' ] = array(
			'formElement' => array(
				'type' => 'hr',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Note: Changing featured image does not update the post modified date.', $this->getPluginTextDomain() )
			) 
		);


		$metaData[ 'FeaturedImageBGImage' ] = array(
			'description' => esc_html__( 'Set Featured Image:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'button_image',
				'default' => '',
				'value'	=> 0,
				'params' => 'Select Image(s)',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Image to use for the featured image.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Set Featured Image', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Image to use for the featured image. If multiple images are selected, one will be chosen at random.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'FeaturedImageBGImageCondition' ] = array(
			'description' => esc_html__( 'Set Featured Image Condition:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'When to set the Featured Image.', $this->getPluginTextDomain() ),
				'value' => array(
					'Always' => 'Always',
					'Only if missing' => 'Only if missing',
				),
				'default' => 'Always',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Set Featured Image Condition', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Allows you to determine if the Featured Image from above will always be set or only if the post is missing a featured image.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'FeaturedCreateHR' ] = array(
			'formElement' => array(
				'type' => 'hr',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Below options are all for Create Image From Title.', $this->getPluginTextDomain() )
			) 
		);

		$metaData[ 'CreateImageFromTitle' ] = array(
			'description' => esc_html__( 'Create Image From Title:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Creates a Featured Image from the post title using the settings below.', $this->getPluginTextDomain() ),
				'value' => array(
					'No' => 'No',
					'Yes' => 'Yes',
					'Only if missing' => 'Only if missing',
				),
				'default' => 'No',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Create Image From Title', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Creates the Featured Image from the Post Title using the settings below. If a featured image already exists, it will be used for the background. The created image filename will be based on the post title automatically.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageHeaderText' ] = array(
			'description' => esc_html__( 'Image Header Text:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'default' => '',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Any text that will appear above the post title in the image.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Image Header Text', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Any text that will appear above the post title in the image.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageFooterText' ] = array(
			'description' => esc_html__( 'Image Footer Text:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'default' => '',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Any text that will appear below the post title in the image.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Image Footer Text', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Any text that will appear below the post title in the image.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageWidth' ] = array(
			'description' => esc_html__( 'Image Width in Pixels:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'default' => get_option( 'thumbnail_size_w'),
				'params' => 'max=10000 min=' . get_option( 'thumbnail_size_w') . ' step=1',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Width of the created featured image.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Featured Image Width', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'The default width is taken from the Settings -> Media Width for the Thumbnail image size. It cannot be lower, but you can specify a larger width here.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageHeight' ] = array(
			'description' => esc_html__( 'Image Height in Pixels:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => get_option( 'thumbnail_size_h'),
				'params' => 'max=10000 min=' . get_option( 'thumbnail_size_h') . ' step=1',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Height of the created featured image.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Featured Image Height', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'The default height is taken from the Settings -> Media Height for the Thumbnail size. It cannot be lower, but you can specify a larger height here.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImagePadding' ] = array(
			'description' => esc_html__( 'Image Padding:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => 10,
				'params' => 'max=1000 min=1 step=1',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Image padding in pixels.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Featured Image Padding', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Image padding in pixels.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageBGColor' ] = array(
			'description' => esc_html__( 'Image Background Hex Color:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'default' => '#ffffff',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Hex color for the background of the created featured image.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Image Background Hex Color', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Hex color for the background of the created featured image.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageTextColor' ] = array(
			'description' => esc_html__( 'Image Text Hex Color:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'default' => '#000000',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Hex color for the text in the created image.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Image Text Hex Color', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Hex color for the text in the created image.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$fontDir = $this->getPluginDir() . 'fonts' . DIRECTORY_SEPARATOR;
		$files = scandir( $fontDir );
		if ( ! empty( $files ) ) {
			$files = array_diff( $files, array( '.', '..' ) );
		}
		else if ( false === $files ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error - scandir of [%s] returned false.', $this->getPluginTextDomain() ), $fontDir ) );
			$files = array();
		}
		$wpUploadData = wp_upload_dir();
		$userDir = $wpUploadData['basedir'] . DIRECTORY_SEPARATOR;
		$userFiles = scandir( $userDir );
		if ( ! empty( $userFiles ) ) {
			$userFiles = array_diff( $userFiles, array( '.', '..' ) );
		}
		else if ( false === $userFiles ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error - scandir of [%s] returned false.', $this->getPluginTextDomain() ), $userDir ) );
			$userFiles = array();
		}
		$files = array_merge( $files, $userFiles );
		$fontTypes = array();
		foreach( $files as $file ) {
			if ( false !== stripos( $file, '.ttf' ) ) {
				$fontTypes[$file] = $file;
			}
		}
		
		$metaData[ 'FeaturedImageFontType' ] = array(
			'description' => esc_html__( 'Image Text Font Type:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'The font type to use.', $this->getPluginTextDomain() ),
				'value' => $fontTypes,
				'default' => '',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Image Text Font Type', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Selections are also taken from the .ttf files in the base WordPress uploads directory (default is wp-content/uploads). Add more fonts as desired.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageFontSize' ] = array(
			'description' => esc_html__( 'Image Text Font Size:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '22',
				'params' => 'max=100 min=1 step=1',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Font size for the text. If the desired size would not fix the image, the font size will be reduced until it does.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Image Text Font Size', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Font size for the text. If the desired size would not fix the image, the font size will be reduced until it does.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageTextTransform' ] = array(
			'description' => esc_html__( 'Image Text Transform:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Any transformation to do on the text.', $this->getPluginTextDomain() ),
				'value' => array(
					'none' => esc_html__( 'none', $this->getPluginTextDomain() ),
					'uppercase' => esc_html__( 'ALL UPPERCASE', $this->getPluginTextDomain() ),
					'lowercase' => esc_html__( 'all lowercase', $this->getPluginTextDomain() ),
					'ucwords' => esc_html__( 'Capitalize Each Word', $this->getPluginTextDomain() )
				),
				'default' => '',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Image Text Transform', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Any transformation to do on the text.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageTextEffect' ] = array(
			'description' => esc_html__( 'Image Text Effect:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Any effects to apply to the text.', $this->getPluginTextDomain() ),
				'value' => array(
					'none' => esc_html__( 'none', $this->getPluginTextDomain() ),
					'drop-shadow' => esc_html__( 'drop shadow', $this->getPluginTextDomain() ),
					'stroke' => esc_html__( 'stroke', $this->getPluginTextDomain() )
				),
				'default' => '',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Image Text Effect', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Drop Shadow places a shadow under and to the side of the text 2 pixels in size.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Long Shadow places a shadow under and to the side of the text that has the same pixel size as the character.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Stroke places a shadow all around the side of the text (like a border) 2 pixels in size.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'In each case the shadow color is determined automatically (black or white) by determining the HSL of the text color.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$encodings = mb_list_encodings();
		$encodings = array_combine( $encodings, $encodings );
		$metaData[ 'FeaturedImageTextEncoding' ] = array(
			'description' => esc_html__( 'Image Text Encoding:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'The text encoding to use.', $this->getPluginTextDomain() ),
				'value' => $encodings,
				'default' => 'UTF-8',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Image Text Encoding', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'The encoding to use for the text. If the font chosen does not support the ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">%s</a> ', esc_url( 'https://www.quackit.com/character_sets/what_is_a_character_set.cfm'), 'encoding' ) . esc_html__( ', those characters will appear strange. UTF-8 (default) will support most languages, but you might want to use other encoding for character sets like Cyrillic or Arabic, and the font will need to support the encoding. This ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">%s</a> ', esc_url( 'https://www.fileformat.info/info/unicode/char/search.htm'), 'website' ) . esc_html__( 'should help.', $this->getPluginTextDomain() ). '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageMaxTextLength' ] = array(
			'description' => esc_html__( 'Image Max Text Length:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '255',
				'params' => 'max=1000 min=1 step=1',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Post titles longer than this length will be trimmed.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Image Max Text Length', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Post titles longer than this length will be trimmed.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageLineSpacing' ] = array(
			'description' => esc_html__( 'Image Text Line Spacing:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '1.5',
				'params' => 'max=10 min=1 step=0.1',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Line spacing for each line of text.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Image Text Line Spacing', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This determines how much height should be used for each line of text. A value of 1 means no line spacing, a value of 1.5 means 150% of the text height, and a value of 2 means double the text height, etc.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'FeaturedImageFileType' ] = array(
			'description' => esc_html__( 'Image Text File Type:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'The file type to save the image as.', $this->getPluginTextDomain() ),
				'value' => array(
					'png' => esc_html__( 'png', $this->getPluginTextDomain() ),
					'jpg' => esc_html__( 'jpg', $this->getPluginTextDomain() )
				),
				'default' => '',
				'tab' => 'Featured Image',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Image Text File Type', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'The file type to save the image as. PNG is best for text only images. Consider JPG if using another image for the background.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		/* --- POST RECYCLING --- */
		
		$qOrder[ 'Daily' ]  = esc_html__( 'Daily Schedule', $this->getPluginTextDomain() );
		$qOrder[ 'Weekly' ]  = esc_html__( 'Weekly Schedule', $this->getPluginTextDomain() );
		$qOrder[ 'Monthly' ]  = esc_html__( 'Monthly Schedule', $this->getPluginTextDomain() );
		$qOrder[ 'OTD' ]  = esc_html__( 'On This Day', $this->getPluginTextDomain() );
		$qOrder[ 'None' ] = esc_html__( 'None', $this->getPluginTextDomain() );
		unset( $qOrder[ 'Newest' ] );
		$metaData[ 'RecycleMode' ] = array(
			'description' => esc_html__( 'Recycle Mode:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Will use all the Post Conditions to find a published post to recycle again as a new post.', $this->getPluginTextDomain() ),
				'value' => $qOrder,
				'default' => 'None',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Recycle Mode', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'All \'Post Conditions\' will be used in Recycle Mode to query for posts, except Post Statuses will be \'publish\', and Recycle Ages will override Post Ages.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'None - Do not recycle content.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Comment Count - The published posts with the most comments will be published again as new.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Oldest Post - The oldest published posts will be published again as new.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Random Posts - Random posts will be published again as new.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Search Relevance - Published posts most matching the Keyword Search will be published again as new.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Daily - Will only recycle posts published today, in order oldest to newest.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Weekly - Will only recycle posts published on the same matching weekday, such as Monday, Tuesday, etc. in order oldest to newest.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Monthly - Will only recycle posts published on the same matching day of the month, such as the 5th, 15th, etc. in order oldest to newest.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'On This Day - Will only recycle posts published the same matching day and month of the year, such as Dec 25th in order oldest to newest.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'RecycleMethod' ] = array(
			'description' => esc_html__( 'Recycle Method:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'How to perform recycling on the selected posts.', $this->getPluginTextDomain() ),
				'value' => array(
					'Republish' => esc_html__( 'Republish', $this->getPluginTextDomain() ),
					'Clone' => esc_html__( 'Clone', $this->getPluginTextDomain() ),
					'Replace' => esc_html__( 'Replace', $this->getPluginTextDomain() )
				),
				'default' => '',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Recycle Method', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Republish - The original post is republished with the current date. This is the default method suitable for most setups.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Clone - A new post is created from the original post and published. The original post remains unchanged, so this could be duplicate content.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Replace - A new post is created from the original post and published. The original post is sent to Trash.', $this->getPluginTextDomain() ) .'</li></ul>' 
			) 
		);
		
		$metaData[ 'MinRecycleAge' ] = array(
			'description' => esc_html__( 'Minimum Recycle Age:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '0',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Posts must be older than this number of days to be recyclable. Enter 0 or leave blank to ignore.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Minimum Post Age', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Posts must be older than this number of days to be recyclable. Enter 0 or leave blank to ignore. Also accepts any ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">%s</a> ', esc_url( 'https://www.php.net/manual/en/function.strtotime.php'), 'strtotime' ) . esc_html__( 'date format', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'MaxRecycleAge' ] = array(
			'description' => esc_html__( 'Maximum Recycle Age:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '0',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Posts must be newer than this number of days to be recyclable. Enter 0 or leave blank to ignore.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Maximum Post Age', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Posts must be older than this number of days to be recyclable. Enter 0 or leave blank to ignore. Also accepts any ', $this->getPluginTextDomain() ) . sprintf( '<a target="_blank" href="%s">%s</a> ', esc_url( 'https://www.php.net/manual/en/function.strtotime.php'), 'strtotime' ) . esc_html__( 'date format', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'OriginalPublish' ] = array(
			'description' => esc_html__( 'Original Publish Note:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Adds an \'originally published\' note to the post content.', $this->getPluginTextDomain() ),
				'value' => array(
					'None' => esc_html__( 'None', $this->getPluginTextDomain() ),
					'Prepend' => esc_html__( 'Prepend', $this->getPluginTextDomain() ),
					'Append' => esc_html__( 'Append', $this->getPluginTextDomain() ) 
				),
				'default' => 'None',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'params' => 'class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Original Publish Note', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'When first recycled, a note with the original publish date can be added.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Prepend - appears before the post content.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Append - appears after the post content.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$metaData[ 'OriginalPublishText' ] = array(
			'description' => esc_html__( 'Original Publish Text:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'subtext' => '',
				'value' => '',
				'default' => '<p>This post was originally published on %%DATE%%.</p>',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Original Publish Text', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'The text to use for the Original Publish Note. Tags supported:', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( '%%DATE%% - The original publish date.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( '%%AUTHOR%% - The author display name.', $this->getPluginTextDomain() ) . '</li></ul>',
				'fullWidth' => true
			) 
		);

		$metaData[ 'NoRecycleHooks' ]  = array(
			'description' => esc_html__( 'Do not Trigger Publish Hooks?', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( '.', $this->getPluginTextDomain() ),
				'value' => '',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Avoid the transition events that normally occur when a post is republished.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Do not Trigger Publish Hooks?', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Normally, when a post is republished a number of publish hooks are triggered that allow other pieces of code to act on those triggers. For instance, if using a plugin that will send a published post to your social media account.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'You may opt to NOT trigger these published hooks when a post is republished (because they were already triggered on the original publish), so that nothing happens except the post published date gets updated to the current time.', $this->getPluginTextDomain() ) . '</li><li>' . esc_html__( 'Note you can only avoid these when the recycle method is set to Republish.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		$metaData[ 'MaxRecycleTimes' ] = array(
			'description' => esc_html__( 'Maximum Times to Recycle:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '0',
				'params' => 'max=36500 min=0 step=1',
				'subtext' => esc_html__( '.', $this->getPluginTextDomain() ),
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'subtext' => esc_html__( 'Maximum times to recycle each post, if desired. Enter 0 to ignore.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Maximum Times to Recycle', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Enter a number if you wish to limit the number of times each post is recycled.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'PostRecycleNotify' ] = array(
			'description' => esc_html__( 'Notify Post Author:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Will send the post author an email when it is recycled.', $this->getPluginTextDomain() ),
				'value' => '',
				'default' => '',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Notify Post Author', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This will send an email to the post author from the site admin notifying them the post has been recycled.', $this->getPluginTextDomain() ) . '</li></ul>',
			) 
		);

		$metaData[ 'PostRecycleForceActions' ] = array(
			'description' => esc_html__( 'Force Post Actions:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Will also apply Post Actions to recycled posts.', $this->getPluginTextDomain() ),
				'value' => '',
				'default' => '',
				'tab' => 'Post Recycling',
				'useIndex' => 'Rulesets',
				'tooltip' => '<h3>' . esc_html__( 'Force Post Actions', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Normally recycling a post just means updating the datetime so it reappears as new. Checking this option will also force Post Actions to recycled posts.', $this->getPluginTextDomain() ) . '</li></ul>',
			) 
		);

		/* --- LOGS --- */
		
		$metaData[ 'LogSize' ]   = array(
			'description' => esc_html__( 'Max Log Size:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '1000',
				'tab' => 'Logs',
				'subtext' => esc_html__( 'Max number of log messages to save.', $this->getPluginTextDomain() ),
				'params' => 'size="8"' 
			) 
		);
		$metaData[ 'DebugMode' ] = array(
			'description' => esc_html__( 'Debug Mode?', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Will log internal messages to help trace any issues.', $this->getPluginTextDomain() ),
				'value' => '',
				'tab' => 'Logs' 
			) 
		);
		
		/* --- LICENSE --- */
		
		$metaData[ 'UserName' ]     = array(
			'description' => esc_html__( 'Envato Username:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'License',
				'subtext' => esc_html__( 'Your username at the', $this->getPluginTextDomain() ) . sprintf( ' <a target="_blank" href="%s">%s</a>', esc_url( 'https://codecanyon.net/author_dashboard' ), esc_html__( 'Envato network', $this->getPluginTextDomain() ) ) . esc_html( ', case sensitive.', $this->getPluginTextDomain() ) 
			) 
		);
		$metaData[ 'PurchaseCode' ] = array(
			'description' => esc_html__( 'Envato Purchase Code:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'License',
				'subtext' => esc_html__( 'Find your purchase code on your', $this->getPluginTextDomain() ) . sprintf( ' <a target="_blank" href="%s">%s</a>', esc_url( 'https://codecanyon.net/downloads' ), esc_html__( 'downloads', $this->getPluginTextDomain() ) ) . esc_html( ' page.', $this->getPluginTextDomain() ),
				'params' => 'size="40"' 
			) 
		);
		$metaData[ 'Support' ]      = array(
			'description' => esc_html__( 'Premium Support Until:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'text',
				'value' => '',
				'tab' => 'License',
				'subtext' => sprintf( '<a target="_blank" href="%s">%s</a>. ', esc_url( 'https://codecanyon.net/page/item_support_policy' ), esc_html__( 'Item support policy', $this->getPluginTextDomain() ) ) . esc_html__( 'Updates are always included.', $this->getPluginTextDomain() ),
				'params' => 'readonly size="40"' 
			) 
		);
		$metaData[ 'Beta' ]         = array(
			'description' => esc_html__( 'Check for Beta updates?', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'checkbox',
				'subtext' => esc_html__( 'Beta updates are used to test bugfixes and new features before official release and may not be stable.', $this->getPluginTextDomain() ),
				'value' => '',
				'tab' => 'License' 
			) 
		);
		
		/* --- ADVANCED --- */
		
		$metaData[ 'SelectMaxEntries' ] = array(
			'description' => esc_html__( 'Select Max Entries:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '1000',
				'params' => 'max=10000 min=0 step=1',
				'tab' => 'Advanced',
				'subtext' => esc_html__( 'Max number of entries in Ruleset select boxes, such as taxonomies/authors, to avoid out of memory issues.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Select Max Entries', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Sets a limit since WordPress functions that retrieve users and taxonomies may run out of memory on large sites.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		$post_taxonomies                = array_diff( get_taxonomies(), array(
			'link_category',
			'nav_menu',
			'post_format' 
		) );
		$metaData[ 'UseTaxonomies' ]    = array(
			'description' => esc_html__( 'Use Taxonomies:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'select',
				'subtext' => esc_html__( 'Which taxonomies to use in Rulesets. Some plugins/themes have custom taxonomies that may or may not be important.', $this->getPluginTextDomain() ),
				'value' => $post_taxonomies,
				'default' => array(
					'category',
					'post_tag' 
				),
				'tab' => 'Advanced',
				'params' => 'multiple size="5" class="wapSelect"',
				'tooltip' => '<h3>' . esc_html__( 'Use Taxonomies', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'Selected entries will be displayed on the Rulesets tab after being saved.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);
		
		$metaData[ 'SleepDelay' ] = array(
			'description' => esc_html__( 'Delay Between Posts:', $this->getPluginTextDomain() ),
			'formElement' => array(
				'type' => 'number',
				'value' => '1',
				'params' => 'max=10000 min=0 step=1',
				'tab' => 'Advanced',
				'subtext' => esc_html__( 'In seconds. This can be used to make sure each post has a unique timestamp.', $this->getPluginTextDomain() ),
				'tooltip' => '<h3>' . esc_html__( 'Delay Between Posts', $this->getPluginTextDomain() ) . '</h3><ul><li>' . esc_html__( 'This can be used to make sure each post has a unique timestamp when modified in bulk. Otherwise if multiple posts have the same timestamp there is no guarantee WordPress will return the same order each time if sorting by date.', $this->getPluginTextDomain() ) . '</li></ul>' 
			) 
		);

		return $metaData;
	}
	
	/**
	 * @return any data to show above the settings
	 */
	public function getSettingsHeader() {
		$support = $this->getOption( 'Support' );
		$expired = false;
		if ( $support ) {
			$s = new \DateTime();
			try {
				$now = new \DateTime( $support );
			}
			catch ( Exception $e ) {
				echo htmlentities( $e->getMessage() );
				exit( 1 );
			}
			$expired = $now <= $s ? true : false;
		}
?>
<div>
        <a target='_blank' href="//www.superblogme.com/wp-auto-poster-documentation/" class="button-primary"><?php
		esc_html_e( ' Documentation', $this->getPluginTextDomain() );
?></a>
<?php
		if ( true === $expired ):
?>
        <a target='_blank' href="http://1.envato.market/Ym9bq" class="button-secondary"><?php
			esc_html_e( 'Renew Support', $this->getPluginTextDomain() );
?></a>
<?php
		else:
?>
        <a target='_blank' href="//www.superblogme.com/contact-me/" class="button-primary"><?php
			esc_html_e( 'Support', $this->getPluginTextDomain() );
?></a>
<?php
		endif;
?>
        <a target='_blank' href="//codecanyon.net/downloads" class="button-primary"><?php
		esc_html_e( 'Rate & Review', $this->getPluginTextDomain() );
?></a>
        <a target='_blank' href="//codecanyon.net/licenses/faq" class="button-primary"><?php
		esc_html_e( 'License FAQ', $this->getPluginTextDomain() );
?></a>
</div>
<p><strong>
<?php
		esc_html_e( 'Automatically update your WordPress site on a flexible schedule with intelligent conditions you fully control.', $this->getPluginTextDomain() );
?>
<br>
<?php
		esc_html_e( 'Plus, it can recycle old content as new again automatically!', $this->getPluginTextDomain() );
?>
</strong></p>
<?php
		if ( !empty( $support ) && true === $expired )
			echo "<i><span class='wp-auto-poster-support-expired'>" . esc_html__( 'Premium support expired on: ', $this->getPluginTextDomain() ) . htmlentities( $support ) . "</span></i>";
		else if ( !empty( $support ) && false === $expired )
			echo "<i>" . esc_html__( 'Premium support expires on: ', $this->getPluginTextDomain() ) . htmlentities( $support ) . "</i>";
	}
	
	protected function initOptions() {
		$options = $this->getOptionMetaData();
		$options = apply_filters( 'wap_init_options', $options );
		if ( $options )
			$this->saveSettings( $options );
	}
	
	public function getPluginTextDomain() {
		return 'wp-auto-poster';
	}
	
	public function getPluginDisplayName() {
		return esc_html__( 'WP Auto Poster', $this->getPluginTextDomain() );
	}
	
	protected function getMainPluginFileName() {
		return 'wp-auto-poster.php';
	}
	
	
	/**
	 * Enqueue admin scripts
	 * @return void
	 */
	public function addAdminScripts() {
		// Example adding a script & style just for the options administration page
		if ( strpos( $_SERVER[ 'REQUEST_URI' ], $this->getSettingsSlug() ) !== false ) {
			wp_enqueue_media();
			wp_enqueue_script( $this->getPluginTextDomain(), $this->GetPluginUrl() . '/js/image-select.js', array( 'jquery' ), $this->getVersion() );
		}
	}

	/**
	 * Enqueue admin styles
	 * @return void
	 */
	public function addAdminStyles() {
		// Example adding a script & style just for the options administration page
		if ( strpos( $_SERVER[ 'REQUEST_URI' ], $this->getSettingsSlug() ) !== false ) {
			wp_enqueue_style( $this->getPluginTextDomain(), $this->getPluginUrl() . '/css/wp-auto-poster.css', array(), $this->getVersion() );
			wp_style_add_data( $this->getPluginTextDomain(), 'rtl', 'replace' );
		}
	}
	
	/**
	 * Add a message to be shown on the admin screen when 'admin_notices' action hook triggered.
	 * @param $msg string text to display
	 */
	protected function addAdminNotice( $msg ) {
		$notices = get_transient( $this->getPluginTextDomain() . '-admin-notices' );
		if ( empty( $notices ) )
			$notices = array(
				 $msg 
			);
		else
			$notices[] = $msg;
		set_transient( $this->getPluginTextDomain() . '-admin-notices', $notices, HOUR_IN_SECONDS );
	}
	
	/**
	 * Show admin messages saved in transient when 'admin_notices' action hook triggered.
	 * @return void
	 */
	public function showAdminNotices() {
		$notices = get_transient( $this->getPluginTextDomain() . '-admin-notices' );
		if ( empty( $notices ) )
			return;
		foreach ( $notices as $notice ) {
			echo wp_kses_post( $notice );
		}
		delete_transient( $this->getPluginTextDomain() . '-admin-notices' );
	}
	
	/**
	 * Initialize plugin actions and filters
	 * @return void
	 */
	public function addActionsAndFilters() {
		
		// Add options administration page
		
		// Add to Settings sub menu:
		add_action( 'admin_menu', array(
			 $this,
			'addSettingsSubMenuPage' 
		) );
		
		add_action( 'admin_enqueue_scripts', array(
			 $this,
			'addAdminScripts' 
		) );
		
		add_action( 'admin_print_styles', array(
			 $this,
			'addAdminStyles' 
		) );
		
		// add Settings shortcut on Plugin page
		$plugin = plugin_basename( $this->getPluginDir() . $this->getMainPluginFileName() );
		add_filter( 'plugin_action_links_' . $plugin, array(
			 $this,
			'addSettingsPluginItems' 
		) );

		// check for special upgrade events
		add_action( 'admin_init', array(
			 $this,
			'checkUpdateSpecial' 
		) );
	
		// Show any plugin admin messages
		add_action( 'admin_notices', array(
			 $this,
			'showAdminNotices' 
		) );
		
		// Check for custom server updates
		add_action( 'plugins_loaded', array(
			 $this,
			'checkForUpdate' 
		) );
		
		// Add custom frequencies from enabled Rulesets to cron schedule
		add_filter( 'cron_schedules', array(
			 $this,
			'addToCronSchedules' 
		), 111, 1 );
		
		// Check for ruleset restart on post publish
		add_action( 'transition_post_status', array(
			 $this,
			'restartRulesetEvent' 
		), 10, 3 );

		// Check for admin action on admin page
		add_action( 'admin_init', array(
			 $this,
			'checkAction' 
		), 10, 1 );

		// Check for custom search modifications:
		add_filter( 'posts_search', array(
			 $this,
			'wapCustomSearch' 
		), 500, 2 );

		// Check for transition status change
		add_action( 'transition_post_status', array(
			 $this,
			'checkTransitionTrigger' 
		), 20, 3 );

		// Check for specific ajax calls
		add_action( 'wp_ajax_wap_get_image', array(
			 $this,
			'getImage' 
		) );

		// Plugin specific hooks -----------------
		add_filter( $this->getPluginTextDomain() . '-createFormElement', array(
			 $this,
			'customFormElement' 
		), 10, 2 );
		add_action( $this->getPluginTextDomain() . '-tab-header', array(
			 $this,
			'tabHeader' 
		), 10, 1 );
		add_action( $this->getPluginTextDomain() . '-tab-footer', array(
			 $this,
			'tabFooter' 
		), 10, 1 );
		add_action( $this->getPluginTextDomain() . '-saved', array(
			 $this,
			'verifyCode' 
		) );

		// Main action hook for all enabled Rulesets
		add_action( $this->getPluginTextDomain(), array(
			 $this,
			'processRuleset' 
		), 10, 1 );


		$support = $this->getOption( 'Support' );
		if ( empty( $support ) ) {
			$plugin = $this->getPluginTextDomain() . '/' . $this->getMainPluginFileName();
        		add_action("after_plugin_row_{$plugin}", function( $plugin_file, $plugin_data, $status ) {
				echo '<tr class="plugin-update-tr active" id="wp-auto-poster-update" data-slug="wp-auto-poster" data-plugin="wp-auto-poster/wp-auto-poster.php"><td colspan="4" class="plugin-update colspanchange"><div class="update-message notice notice-warning inline notice-alt"><p>'
				. esc_html__( 'This plugin is not registered. Automatic updates are disabled. Please enter your purchase code in the \'License\' tab in Settings to unlock. ', 'video-link-checker')
				. '</p></div></td></tr>';
        		}, 10, 3 );
		}
	}
	
	/**
	 * Check WAP version upgrade for special events
	 * @return void
	 */
	public function checkUpdateSpecial() {

		if ( ! $this->isInstalledCodeAnUpgrade() )
			return;

		// WordPress 6.0 changed how named parameters work. Code fix.
		if ( $this->isSavedVersionLessThanEqual( '1.7' ) ) {
			$this->restartRulesets();
			$this->saveInstalledVersion();
		}

		// replaced setting 'FeaturedImageOnPublish' with 'TriggerOnTransition'
		if ( $this->isSavedVersionLessThanEqual( '2.2' ) ) {
			$save = false;
			$rulesets = $this->getOption( 'Rulesets', array ());
			foreach ( $rulesets as $ruleset ) {
				$oldTrigger = $this->getRulesetOption( $ruleset['ID'], 'FeaturedImageOnPublish', 'off' );
				// if using old setting, set new setting
				if ( 'on' === $oldTrigger ) {
					$this->setRulesetOption( $ruleset['ID'], 'TriggerOnTransition', 'on' );
					$save = true;
				}
			}
			if ( true === $save ) {
				$this->restartRulesets();
				$this->saveInstalledVersion();
			}
		}
	}
	
		
	/**
	 * Check private server for updates
	 * This happens whenever WP checks for updates, default is every 12 hours
	 * @return void
	 */
	public function checkForUpdate() {
		
		$username = $this->getOption( 'UserName' );
		$code     = $this->getOption( 'PurchaseCode' );
		$support  = $this->getOption( 'Support' );
		$beta     = $this->getOption( 'Beta' );
		
		if ( !$username || !$code || !$support )
			return;
		
		$url = $beta === 'on' ? "https://www.superblogme.com/test-update-server/?action=get_metadata&slug=" : "https://www.superblogme.com/wp-update-server/?action=get_metadata&slug=";
		require_once $this->getPluginDir() . '/plugin-updates/plugin-update-checker.php';
		$updateChecker = Puc_v4_Factory::buildUpdateChecker( 
			$url . $this->getPluginTextDomain(), 
			$this->getPluginDir() . $this->getMainPluginFileName(), 
			$this->getPluginTextDomain() 
		);
		$updateChecker->addQueryArgFilter( array(
			 $this,
			'PUC_callback' 
		) );
	}
	
	/**
	 * Add our own arguments onto the PUC
	 * @param the array of default args from the PUC
	 * @return the array with custom data
	 */
	public function PUC_callback( $query_args ) {
		if ( empty( $query_args ) )
			$query_args = array();
		$query_args[ 'license' ] = $this->getOption( 'UserName' );
		$query_args[ 'code' ]    = $this->getOption( 'PurchaseCode' );
		return $query_args;
	}
	
	/**
	 * Get a ruleset by ID
	 * @param ID number 
	 * @return the array with the Ruleset
	 */
	protected function getRuleset( $id ) {
		
		if ( !$this->options )
			$this->loadOptions();
		foreach ( $this->options[ 'Rulesets' ] as $ruleset ) {
			if ( (int) $ruleset[ 'ID' ] === (int) $id ) {
				return $ruleset;
			}
		}
		// should not happen:
		return null;
	}
	
	/**
	 * Get a ruleset option by ID, optionName
	 * @param ID number of the ID OR full ruleset as an array
	 * @param optionName string 
	 * @return the optionName value from the matching ruleset
	 */
	protected function getRulesetOption( $id, $optionName, $default = null ) {
		$ruleset = is_array( $id ) ? $id : $this->getRuleset( $id );
		if ( $ruleset && isset( $ruleset[ $optionName ] ) ) {
			return $ruleset[ $optionName ];
		}
		else if ( $ruleset ) {
			return $default;
		}
		return $default;
	}
	
	/**
	 * Set a ruleset option by ID, optionName
	 * @param ID number of the ID
	 * @param optionName string 
	 * @return void
	 */
	protected function setRulesetOption( $id, $optionName, $value ) {
		if ( !$this->options )
			$this->loadOptions();
		foreach ( $this->options[ 'Rulesets' ] as $ruleset ) {
			if ( $ruleset[ 'ID' ] === $id ) {
				$ruleset[ $optionName ] = $value;
				return;
			}
		}
	}
	
	/**
	 * Customize FormElements dynamically depending on certain rules.
	 * @param $aOptionMeta array of this formElement's data
	 * @param $aOptionKey string the name of this formElement
	 * @return array of this formElement's data
	 */
	public function customFormElement( $aOptionMeta, $aOptionKey ) {
		
		
		// set value on fly
		if ( $aOptionKey === 'Status' ) {
			$id      = $this->getVarID();
			$ruleset = empty( $id ) ? 0 : $this->getRuleset( $id );
			
			if ( empty( $ruleset ) ) {
			} else if ( $ruleset[ 'Status' ] === 'disabled' ) {
				$aOptionMeta[ 'formElement' ][ 'subtext' ] = ' 
<input type="submit" class="button-primary" name="wapAction"
    value="' . esc_attr__( "Enable Ruleset", $this->getPluginTextDomain() ) . '"/>
<input type="submit" class="button-primary" name="wapAction"
    value="' . esc_attr__( "Run Once", $this->getPluginTextDomain() ) . '"/>
				';
			} else {
				$aOptionMeta[ 'formElement' ][ 'subtext' ] = ' 
<input type="submit" class="button-primary" name="wapAction"
    value="' . esc_attr__( "Disable Ruleset", $this->getPluginTextDomain() ) . '"/>
<input type="submit" class="button-primary" name="wapAction"
    value="' . esc_attr__( "Run Once", $this->getPluginTextDomain() ) . '"/>
				';
			}
			if ( ! empty( $ruleset ) ) {
				$aOptionMeta[ 'formElement' ][ 'subtext' ] .= ' 
<input type="submit" class="button-primary" name="wapAction"
    value="' . esc_attr__( "Export Ruleset", $this->getPluginTextDomain() ) . '"/>
				';
				$aOptionMeta[ 'formElement' ][ 'subtext' ] .= ' 
<input type="submit" class="button-primary" name="wapAction"
    value="' . esc_attr__( "Delete Ruleset", $this->getPluginTextDomain() ) . '"/>
				';
			}
		}
		
		// set value on fly
		else if ( $aOptionKey === 'Support' ) {
			// attach css class depending on if support has expired.
			$support = $this->getOption( 'Support' );
			if ( $support ) {
				$s            = new \DateTime();
				$now          = new \DateTime( $support );
				$supportClass = $now <= $s ? 'wp-auto-poster-support-expired' : 'wp-auto-poster-support-valid';
				$supportClass .= ' wp-auto-poster-input-transparent';
				$aOptionMeta[ 'formElement' ][ 'params' ] .= ' class="' . $supportClass . '"';
			}
		}
		
		// set the hidden input 'ID' dynamically
		else if ( $aOptionKey === 'ID' ) {
			$nextID                                  = $this->getOption( 'nextID', 1 );
			$aOptionMeta[ 'formElement' ][ 'value' ] = $nextID;
		}
		
		return $aOptionMeta;
	}
	
	/**
	 * Verify license data
	 * @return void
	 */
	public function verifyCode() {
		
		$support = $this->getOption( 'Support' );
		
		$username = $this->getOption( 'UserName' );
		if ( empty( $support ) && ! $username ) {
			echo "<div class='error'><p>" . esc_html__( $this->getPluginDisplayName() . ' License: missing Username.', $this->getPluginTextDomain() ) . "</p></div>";
			return;
		}
		$code = $this->getOption( 'PurchaseCode' );
		if ( empty( $support ) && !$code ) {
			echo "<div class='error'><p>" . esc_html__( $this->getPluginDisplayName() . ' License: missing Purchase Code.', $this->getPluginTextDomain() ) . "</p></div>";
			return;
		}

		// sometimes a user will enter a space in their username...
		$username = str_replace( ' ', '', $username );
		
		// check transient first - no need to re-verify license every save.
		if ( ! empty( $support ) ) {

			$value = get_transient( $this->getPluginTextDomain() . '-license-buffer' );
			if ( false !== $value )
				return;
		}
		
		// check jph codes
		if ( substr( $code, 0, 3 ) === 'jph' ) {
			$verify_url = 'http://www.superblogme.com/verify/wap.php?user=' . $username . '&code=' . $code . '&site=' . site_url();
			$response   = wp_remote_get( $verify_url );
			if ( is_wp_error( $response ) ) {
				if ( !$support ) {
					$msg = sprintf( esc_html__( 'WP error %s in WAP %s - %s', $this->getPluginTextDomain() ), $response->get_error_code(), __FUNCTION__, $response->get_error_message() );
					echo "<div class='error'><p>" . htmlentities( $msg ) . "</p></div>";
				}
				return;
			}
			$msg = json_decode( $response[ 'body' ] );
			if ( !$support && !$msg ) {
				echo "<div class='error'><p>" . $this->getPluginDisplayName() . " - Could not decode response from server.</p></div>";
				return;
			}
			
			if ( $msg->code === 201 ) {
				$this->updateOption( 'Support', date( DATE_ISO8601, strtotime( '+1 years' ) ) );
				set_transient( $this->getPluginTextDomain() . '-license-buffer', date( DATE_ISO8601, strtotime( '+1 years' ) ), WEEK_IN_SECONDS );
			} else if ( !$support ) {
				echo "<div class='error'><p>" . $this->getPluginDisplayName() . " - " . $msg->msg . "</p></div>";
			}
			return;
		}
		
		$headers    = array(
			'Authorization' => 'Bearer xaWVKX1aqpxxpk1UdOIC2Z8b0TUe77dw' 
		);
		$verify_url = 'https://api.envato.com/v2/market/author/sale?code=' . $code;
		
		$response = wp_remote_get( $verify_url, array(
			'headers' => $headers,
			'user-agent' => $this->getPluginDisplayName() 
		) );
		
		if ( is_wp_error( $response ) ) {
			if ( !$support )
				printf( esc_html__( 'WP error %s in WAP %s - %s', $this->getPluginTextDomain() ), $response->get_error_code(), __FUNCTION__, $response->get_error_message() );
			return;
		}
		
		$verify_obj = json_decode( wp_remote_retrieve_body( $response ) );
		
		$verify_obj = apply_filters( $this->getPluginTextDomain() . '-verify', $verify_obj );
		
		// Check for correct verify code
		if ( ( false === $verify_obj ) || !is_object( $verify_obj ) )
			$msg = sprintf( esc_html__( "%s: Could not connect to Envato API server.", $this->getPluginTextDomain() ), $this->getPluginDisplayName() );
		else if ( isset( $verify_obj->error ) && isset( $verify_obj->description ) )
			$msg = sprintf( esc_html__('%s: Envato error: %s', $this->getPluginTextDomain() ), $this->getPluginDisplayName(), $verify_obj->description );
		else if ( !isset( $verify_obj->item ) || !isset( $verify_obj->item->name ) )
			$msg = sprintf( esc_html__( "%s: Envato API did not return item information.", $this->getPluginTextDomain() ), $this->getPluginDisplayName() );
		
		// Check for proper item name
		else if (
		FALSE === stripos( $verify_obj->item->name, 'WP Auto Poster' )
		)
		$msg = sprintf( esc_html__('Envato API states the %s purchase code is for: %s', $this->getPluginTextDomain() ), $this->getPluginDisplayName(), $verify_obj->item->name );
		
		// Check for proper item id
		else if (
		$verify_obj->item->id != 23805990
		)
		$msg = sprintf( esc_html__( "Envato API states the %s purchase code is for another product id.", $this->getPluginTextDomain() ), $this->getPluginDisplayName() );
		
		// Check for username match
		else if ( !empty( $verify_obj->buyer ) && strcmp( $verify_obj->buyer, $username ) )
			$msg = sprintf( esc_html__( "Envato states the %s purchase code is for another username.", $this->getPluginTextDomain() ), $this->getPluginDisplayName() );
		
		else if ( isset( $verify_obj->supported_until ) && ! empty( $verify_obj->supported_until ) ) {
			$this->updateOption( 'Support', $verify_obj->supported_until );
			set_transient( $this->getPluginTextDomain() . '-license-buffer', $verify_obj->supported_until, DAY_IN_SECONDS );
			$msg = null;
		} else
			$msg = sprintf( esc_html__( "%s: Strange. Envato API did not return a support field.", $this->getPluginTextDomain() ), $this->getPluginDisplayName() );
		
		$support = $this->getOption( 'Support' );
		if ( !$support && $msg )
			echo "<div class='error'><p>" . $msg . "</p></div>";
	}
	
	/* Custom functions go here --------------------------------------------- */
	
	/**
	 * Add a message to the log
	 * @param $msg string the message to add
	 */
	protected function addLogMessage( $msg, $echo = true ) {
		if ( !$this->logs )
			$this->logs = get_option( $this->getClassName() . '-logs', array ());
		$class = $this->getPluginTextDomain() . '-log-date';
		if ( false !== stripos( $msg, ' error' ) )
			$class = $this->getPluginTextDomain() . '-log-error';
		else if ( false !== stripos( $msg, 'debug' ) ) {
			if ( $this->getOption( 'DebugMode' ) !== 'on' )
				return;		// do not log debug msgs if Debug Mode not on
			$class = $this->getPluginTextDomain() . '-log-debug';
		}
		
		$this->logs[] = '<span class="' . $class . '">' . date_i18n( 'Y-m-d H:i:s' ) . '</span> - ' . $msg;
		
		$this->saveLogMessages();
		
		// for any messages before pluggable loaded:
		if ( ! function_exists( 'wp_get_current_user' ) ) {
			return false;
		}

		// only echo if on WAP page
		$screen = function_exists( 'get_current_screen' ) ? get_current_screen() : null;
		if ( ! empty( $screen ) && false === stripos( $screen->base, 'WpAutoPoster' ) )
			return;

		if ( $echo && current_user_can( 'moderate_comments' ) ) {
			if ( false !== stripos( $msg, 'error' ) )
				$amsg = "<div class='error'><p>" . $msg . "</p></div>";
			else if ( false !== stripos( $msg, 'warning' ) )
				$amsg = "<div class='notice notice-warning'><p>" . $msg . "</p></div>";
			else if ( false !== stripos( $msg, 'debug' ) )
				$amsg = "<div class='notice notice-warning'><p>" . $msg . "</p></div>";
			else
				$amsg = "<div class='updated'><p>" . $msg . "</p></div>";
			$this->addAdminNotice( $amsg );
		}
		return false;
	}
	
	/**
	 * Save the log messages to option field, if any, keeping it under max log size
	 * @return void
	 */
	protected function saveLogMessages() {
		if ( $this->logs ) {
			$maxSize    = $this->getOption( 'LogSize', 1000 );
			$this->logs = array_slice( $this->logs, -1 * $maxSize );
			update_option( $this->getClassName() . '-logs', $this->logs, false );
		}
	}
	
	/**
	 * Clear the log messages option field
	 * @return void
	 */
	protected function clearLogMessages() {
		if ( ! delete_option( $this->getClassName() . '-logs' ) )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error - %s: function delete_option() returned false.', $this->getPluginTextDomain() ), __FUNCTION__ ) );

		$this->logs = array();
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s complete.', $this->getPluginTextDomain() ), __FUNCTION__ ) );
	}
	
	/**
	 * Show log messages 
	 * @return void
	 */
	protected function showLogMessages() {
		if ( empty( $this->logs ) )
			$this->logs = get_option( $this->getClassName() . '-logs', array () );
		echo '<div id="' . $this->getPluginTextDomain() . '-log-messages">';
		foreach ( $this->logs as $log ) {
			echo wp_kses_post( $log ) . "<br>";
		}
		echo '</div>';
	}
	
	/**
	 * Show Rulesets
	 * @return void
	 */
	protected function showRulesets() {
		if ( !$this->options )
			$this->loadOptions();
		if ( !empty( $this->options[ 'Rulesets' ] ) ) {
			$formatted = date_i18n( "Y-m-d H:i:s", current_time( 'timestamp' ) ) . " " . get_option( 'timezone_string' );
			$timeNow   = sprintf( esc_html__( 'Time now is %s', $this->getPluginTextDomain() ), $formatted );
?>
<h3>
<?php echo '<img width="28" height="28" alt="Saved Rulesets" title="Saved Rulesets" src="' . $this->getPluginUrl() . '/images/wap-saved.png" />'; ?>
 Saved Rulesets
</h3>
<div class="divTable wapTable">
<div class="divTableHeading">
<div class="divTableRow">
<div class="divTableHead"><?php
			esc_html_e( 'ID', $this->getPluginTextDomain() );
?></div>
<div class="divTableHead divTableNoWrap"><?php
			esc_html_e( 'Status', $this->getPluginTextDomain() );
?> <span class='divTableHeadSmall'><?php
			echo htmlentities( $timeNow );
?></span></div>
<div class="divTableHead"><?php
			esc_html_e( 'Name', $this->getPluginTextDomain() );
?></div>
<div class="divTableHead"><?php
			esc_html_e( 'Frequency', $this->getPluginTextDomain() );
?></div>
<div class="divTableHead"><?php
			esc_html_e( 'Action', $this->getPluginTextDomain() );
?></div>
</div>
</div>
<div class="divTableBody">
		<?php
			$editId = $this->getVarID();
			foreach ( $this->options[ 'Rulesets' ] as $ruleset ) {
				if ( !is_array( $ruleset ) )
					continue;
				$id     = isset( $ruleset[ 'ID' ] ) ? (int) $ruleset[ 'ID' ] : '?';
				$active = $editId === $id ? 'divRulesetActive' : '';
				$status = $statusShow = isset( $ruleset[ 'Status' ] ) ? $ruleset[ 'Status' ] : 'disabled';
				if ( $status === 'Enabled' ) {
					$eventTime = $this->getRulesetEventTime( $ruleset[ 'ID' ] );
					if ( $eventTime ) {
						$statusShow .= ' - ' . $eventTime;
					}
				}
				$name       = isset( $ruleset[ 'Name' ] ) ? $ruleset[ 'Name' ] : '';
				$freq       = isset( $ruleset[ 'Frequency' ] ) ? $ruleset[ 'Frequency' ] : '?';
				$btnEditUrl = sprintf( 'admin.php?page=%sSettings&editRuleset=%s', htmlentities( $this->getClassName() ), htmlentities( $id ) );
				$btnEnableUrl = sprintf( 'admin.php?page=%sSettings&enableRuleset=%s', htmlentities( $this->getClassName() ), htmlentities( $id ) );
				$btnDisableUrl = sprintf( 'admin.php?page=%sSettings&disableRuleset=%s', htmlentities( $this->getClassName() ), htmlentities( $id ) );
				$btnDelUrl  = sprintf( 'admin.php?page=%sSettings&delRuleset=%s', htmlentities( $this->getClassName() ), htmlentities( $id ) );
				$btnExportUrl  = sprintf( 'admin.php?page=%sSettings&exportRuleset=%s', htmlentities( $this->getClassName() ), htmlentities( $id ) );
?>
<div class="divTableRow">
<div class="divTableCell <?php
				echo htmlentities( $active );
?>"><?php
				echo htmlentities( $id );
?></div>
<div class="divTableCell divTableNoWrap <?php
				echo htmlentities( $active );
?>"><?php
				if ( $status === 'Enabled' )
					echo '<span class="divRulesetEnabled">';
				echo htmlentities( $statusShow );
				if ( $status === 'Enabled' )
					echo '</span>';
?></div>
<div class="divTableCell divTableFull <?php
				echo htmlentities( $active );
?>">
<a href="<?php
				echo esc_url( $btnEditUrl );	// nonce not needed
?>">
<?php
				echo htmlentities( $name );
?></a></div>
<div class="divTableCell <?php
				echo htmlentities( $active );
?>"><?php
				echo htmlentities( $freq );
?></div>
<div class="divTableCell divTableNoWrap <?php
				echo htmlentities( $active );
?>">

<a href="<?php
				echo esc_url( $btnEditUrl );	// nonce not needed
?>"><img width="16" height="16" alt="Edit this Ruleset" title="Edit this Ruleset" src="<?php
				echo esc_url( $this->getPluginUrl() );
?>/images/list_edit.png"></a>

<?php				if ( $status === 'Enabled' ) { ?>
<a href="<?php
				echo wp_nonce_url( $btnDisableUrl, 'DISABLERULESET=' . htmlentities( $id ), $this->getPluginTextDomain() . '_nonce' );
?>"><img width="16" height="16" alt="Disable this Ruleset" title="Disable this Ruleset" src="<?php
				echo esc_url( $this->getPluginUrl() );
?>/images/disable.png"></a>
<?php } else { ?>

<a href="<?php
				echo wp_nonce_url( $btnEnableUrl, 'ENABLERULESET=' . htmlentities( $id ), $this->getPluginTextDomain() . '_nonce' );
?>"><img width="16" height="16" alt="Enable this Ruleset" title="Enable this Ruleset" src="<?php
				echo esc_url( $this->getPluginUrl() );
?>/images/enable.png"></a>
<?php } ?>

<a href="<?php
				echo wp_nonce_url( $btnExportUrl, 'EXPORTRULESET=' . htmlentities( $id ), $this->getPluginTextDomain() . '_nonce' );
?>"><img width="16" height="16" alt="Export this Ruleset" title="Export this Ruleset" src="<?php
				echo esc_url( $this->getPluginUrl() );
?>/images/export.png"></a>

<a href="<?php
				echo wp_nonce_url( $btnDelUrl, 'DELETERULESET=' . htmlentities( $id ), $this->getPluginTextDomain() . '_nonce' );
?>"><img width="16" height="16" alt="Delete this Ruleset" title="Delete this Ruleset" src="<?php
				echo esc_url( $this->getPluginUrl() );
?>/images/list_remove.png"></a>

</div>
</div>
		<?php
			}
?>
</div>
</div>
		<?php 
			// Noteworthy messages:
			if ( defined( 'DISABLE_WP_CRON' ) ) {
				printf(
				'<small>%s <a target="_blank" href="%s">%s</a> %s</small>',
				esc_html__( 'WP-Cron not detected. Enabled rulesets should trigger when a ', $this->getPluginTextDomain() ),
				esc_url( 'https://www.google.com/search?q=wordpress+server+cron+job' ),
				esc_html__( 'server cron job', $this->getPluginTextDomain() ),
				esc_html__( 'calls wp-cron directly after their scheduled event time.', $this->getPluginTextDomain() )
				);
			}
			else {
				printf(
				'<small><a target="_blank" href="%s">%s</a> %s</small>',
				esc_url( 'https://developer.wordpress.org/plugins/cron/' ),
				esc_html__( 'WP-Cron', $this->getPluginTextDomain() ),
				esc_html__( ' detected. Enabled rulesets should trigger on first page load after their scheduled event time.', $this->getPluginTextDomain() )
				);
			}
			if ( is_plugin_active( 'w3-total-cache/w3-total-cache.php' ) ) {
				printf(
				'<br><small><a target="_blank" href="%s">%s</a> %s</small>',
				esc_url( 'https://wordpress.org/plugins/w3-total-cache/' ),
				esc_html__( 'W3 Total Cache', $this->getPluginTextDomain() ),
				esc_html__( ' detected. Note: If its Object Cache is enabled it will mess up the cron transient.', $this->getPluginTextDomain() )
				);
			}
			if ( is_plugin_active( 'litespeed-cache/litespeed-cache.php' ) ) {
				printf(
				'<br><small><a target="_blank" href="%s">%s</a> %s</small>',
				esc_url( 'https://wordpress.org/plugins/litespeed-cache/' ),
				esc_html__( 'LiteSpeed Cache', $this->getPluginTextDomain() ),
				esc_html__( ' detected. Note: If its Object Cache is enabled it will mess up the cron transient.', $this->getPluginTextDomain() )
				);
			}
		}
?>

		<p>

		<?php
			$id = $this->getVarID();
			if ( $id ) {
?>
			<input type="submit" class="button-primary" name="wapAction"
				formaction="admin.php?page=<?php
				echo htmlentities( $this->getClassName() ) . 'Settings';
?>"
				value="<?php
				esc_html_e( 'New Ruleset', $this->getPluginTextDomain() );
?>"/>
			&nbsp; &nbsp; &nbsp; or &nbsp; &nbsp; &nbsp;
		<?php
			}
?>
		<input type="submit" name="wapAction"
    			value="<?php 
				esc_html_e( "Import Ruleset", $this->getPluginTextDomain() );
?>"/>
		<input type="file" id="wap_import_file" name="wap_import_file" accept=".json">
		</p>
		<hr>
		<?php
	}
	
	/**
	 * Get the global var rulesetID, either from the _POST or _GET environment, depending on action
	 * @return the ID or null
	 */
	protected function getVarID() {
		if ( isset( $_GET[ 'editRuleset' ] ) )
			$id = filter_input( INPUT_GET, 'editRuleset', FILTER_VALIDATE_INT );
		else if ( isset( $_GET[ 'enableRuleset' ] ) )
			$id = filter_input( INPUT_GET, 'enableRuleset', FILTER_VALIDATE_INT );
		else if ( isset( $_GET[ 'disableRuleset' ] ) )
			$id = filter_input( INPUT_GET, 'disableRuleset', FILTER_VALIDATE_INT );
		else if ( isset( $_GET[ 'exportRuleset' ] ) )
			$id = filter_input( INPUT_GET, 'exportRuleset', FILTER_VALIDATE_INT );
		else if ( isset( $_POST[ 'ID' ] ) 
			&& isset( $_POST[ 'wapAction' ] ) 
			&& in_array( $_POST[ 'wapAction' ], array( 'Enable Ruleset', 'Disable Ruleset', 'Run Once', 'Save Changes', 'Delete Ruleset' ) )
		)
			$id = filter_input( INPUT_POST, 'ID', FILTER_VALIDATE_INT );
		else
			$id = null;

		if ( ! $this->getRuleset( $id ) )
			$id = null;

		return $id;
	}
	
	/**
	 * Display custom data at the header of a tab
	 * @param $tabID string sanitized name of this tab (lowercase, hyphens instead of spaces)
	 */
	public function tabHeader( $tabID ) {
		if ( $tabID === 'wp-auto-poster-tab-rulesets' ) {
			$this->showRulesets();
			$id = $this->getVarID();
			if ( $id ) {
				echo '<h3>';
				echo '<img width="28" height="28" alt="Edit Ruleset" title="Edit Ruleset" src="' . $this->getPluginUrl() . '/images/wap-edit.png" />';
				echo ' Edit Ruleset #' . $id;
				echo '</h3>';
			}
			else {
				echo '<h3>';
				echo '<img width="28" height="28" alt="Create new Ruleset" title="Create new Ruleset" src="' . $this->getPluginUrl() . '/images/wap-create.png" />';
				echo ' Create new Ruleset';
				echo '</h3>';
			}
		} else if ( $tabID === 'wp-auto-poster-tab-general' ) {
			printf( esc_html__( '%sThis Ruleset will be checked every time it is triggered by the Event Frequency%sIf it passes the Inclusion/Exclusion dates and Max Posts per Day, WAP will process the Post Conditions%s', $this->getPluginTextDomain() ), '<h4><i>', '<br>', '</i></h4>' );
		} else if ( $tabID === 'wp-auto-poster-tab-post-conditions' ) {
			printf( esc_html__( '%sWordPress function WP_Query will be called with these settings to find any eligible posts to process Post Actions upon%s', $this->getPluginTextDomain() ), '<h4><i>', '</i></h4>' );
		} else if ( $tabID === 'wp-auto-poster-tab-post-actions' ) {
			printf( esc_html__( '%sWhat to modify when eligible posts are found using the Post Conditions%s', $this->getPluginTextDomain() ), '<h4><i>', '</i></h4>' );
		} else if ( $tabID === 'wp-auto-poster-tab-featured-image' ) {
			printf( esc_html__( '%sCan also Modify, Find or Create a featured image for each eligible post%s', $this->getPluginTextDomain() ), '<h4><i>', '</i></h4>' );
		} else if ( $tabID === 'wp-auto-poster-tab-post-recycling' ) {
			printf( esc_html__( '%sIf not enough eligible posts are found when querying Post Conditions, you can optionally recycle content to show something fresh%s', $this->getPluginTextDomain() ), '<h4><i>', '</i></h4>' );
		} else if ( $tabID === 'wp-auto-poster-tab-license' ) {
			printf( esc_html__( '%sLicense info will be sent to the Envato API to verify purchase, and check for updates on author\'s server%s', $this->getPluginTextDomain() ), '<h4><i>', '</i></h4>' );
		}
	}
	
	/**
	 * Display custom data at the footer of a tab
	 * @param $tabID string sanitized name of this tab (lowercase, hyphens instead of spaces)
	 */
	public function tabFooter( $tabID ) {
		
		if ( $tabID === 'wp-auto-poster-tab-logs' ) {
			$this->showLogMessages();
?>
			<p>
			<input type="submit" class="button-primary" name="wapAction"
				value="<?php
			esc_html_e( 'Clear Logs', $this->getPluginTextDomain() );
?>"/>
			</p>
<?php
		}
		
	}
	
	/**
	 * Save new event time in Ruleset subIndex
	 * @param $ruleset array to update
	 */
	private function getRulesetEventTime( $rulesetID, $verifyCron = true ) {
		
		$args     = array(
			(int) $rulesetID 
		);
		$hookname = $this->getPluginTextDomain();
		
		$scheduledtime = wp_next_scheduled( $hookname, $args );
		if ( false === $scheduledtime ) {

			if ( true === $verifyCron ) {
				// not found!? should not happen.
				$this->addLogMessage( sprintf( esc_html__( 'WAP Warning - Could not find cron schedule entry for %s, %s. Likely another plugin overwrote cron entries. Will try rescheduling event.', $this->getPluginTextDomain() ), $hookname, print_r( $args, true ) ) );
				$this->enableRuleset( $rulesetID );
			}
			return null;
		}

		$formatted = date_i18n( "Y-m-d H:i:s", $scheduledtime + ( get_option( 'gmt_offset' ) * HOUR_IN_SECONDS ) ) . " " . get_option( 'timezone_string' );

		return 'Next event after ' . $formatted;
	}
	
	/**
	 * restart all enabled Rulesets
	 */
	protected function restartRulesets() {
		$rulesets = $this->getOption( 'Rulesets', array ());
		foreach ( $rulesets as $ruleset ) {
			if ( !empty( $ruleset[ 'Frequency' ] ) && isset( $ruleset[ 'Status' ] ) && false !== stripos( $ruleset[ 'Status' ], 'Enabled' ) ) {
				$this->enableRuleset( $ruleset[ 'ID' ] );
			}
		}
	}

	/**
	 * Enable ruleset by setting event in cron schedule
	 * @param $rulesetID number of the ruleset to enable
	 * @param $restart boolean whether we are restarting an existing event
	 */
	protected function enableRuleset( $rulesetID = null, $restart = false ) {
		
		$support = $this->getOption( 'Support' );
		if ( empty( $support ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Please enter your License information to Enable Rulesets.', $this->getPluginTextDomain() ) ) );
			return;
		}
		
		if ( !$rulesetID )
			$rulesetID = $this->getVarID();
		
		$ruleset = $this->getRuleset( $rulesetID );
		if ( !$ruleset )
			return;
		
		$args     = array(
			(int) $rulesetID 
		);
		$hookname = $this->getPluginTextDomain();
		wp_clear_scheduled_hook( $hookname, $args );
		
		$customRecurrence = $this->getPluginTextDomain() . '-' . $rulesetID;
		
		if ( empty( $ruleset[ 'Frequency' ] ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - Event Frequency is empty.', $this->getPluginTextDomain() ), $rulesetID ) );
			return;
		}
		
		$delay = true === $restart ? $this->convertTime( $ruleset[ 'Frequency' ], false ) : $this->convertTime( $ruleset[ 'Frequency' ], true );
		$delay = apply_filters( $this->getPluginTextDomain() . '-delay-time', $delay, $ruleset );
		
		$ruleset[ 'Status' ] = sprintf( esc_html__( 'Enabled', $this->getPluginTextDomain() ) );
		$timeInSecs = $this->convertTime( $ruleset[ 'Frequency' ] );
		$timeInSecs = apply_filters( $this->getPluginTextDomain() . '-epoch-time', $timeInSecs, $ruleset );
		$ruleset[ 'timeInSecs' ] = $timeInSecs;

		if ( $delay )
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset frequency %s converted to %s seconds (delay).', $this->getPluginTextDomain() ), __FUNCTION__, $ruleset['Frequency'], $delay, $ruleset['timeInSecs'] ) );
		else
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset frequency %s converted to %s seconds.', $this->getPluginTextDomain() ), __FUNCTION__, $ruleset['Frequency'], $ruleset['timeInSecs'] ) );
		$this->addUpdateSubIndex( 'Rulesets', $ruleset, true );

		$schedules = wp_get_schedules();
		
		if ( !isset( $schedules[ $customRecurrence ] ) )
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Could not find cron schedule entry for %s. Likely another plugin overwrote cron entries. wp_get_schedules returned: %s', $this->getPluginTextDomain() ), $customRecurrence, print_r($schedules,true) ) );
		else if ( wp_next_scheduled( $hookname, $args ) )
			$this->addLogMessage( sprintf( esc_html__( 'Notice - Cron entry already found for %s 
(%s).', $this->getPluginTextDomain() ), $hookname, print_r($args,true) ) );
		else if ( false !== wp_schedule_event( time() + $delay, $customRecurrence, $hookname, $args ) ) {
			if ( true === $restart )
				$this->addLogMessage( sprintf( esc_html__( 'A post has been published outside of WAP. Restart on Publish is checked, so Ruleset #%s event has been restarted with timestamp %d.', $this->getPluginTextDomain() ), $rulesetID, time() + $delay ) );
			else
				$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s has been enabled with hookname [%s], custom schedule: [%s], args: %s, timestamp: %d.', $this->getPluginTextDomain() ), $rulesetID, $hookname, $customRecurrence, print_r($args,true), time() + $delay ) );
			return;	// success!
		} else
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: wp_schedule_event() for ruleset #%s failed because another plugin disallowed this event.', $this->getPluginTextDomain() ), $rulesetID ) );
		
		// if here the event was not scheduled - set Status back to disabled
		$this->disableRuleset( $rulesetID, true );
	}
	
	/**
	 * Disable ruleset by clearing event in cron schedule
	 * @param $rulesetID number of the ruleset to disable
	 */
	protected function disableRuleset( $rulesetID = null, $save = false ) {
		if ( !$rulesetID )
			$rulesetID = $this->getVarID();
		
		$ruleset = $this->getRuleset( $rulesetID );
		if ( !$ruleset )
			return;
		
		// if Ruleset is not enabled just return
		if ( isset( $ruleset[ 'Status' ] ) && false === stripos( $ruleset[ 'Status' ], 'Enabled' ) )
			return;
		
		$args     = array(
			(int) $rulesetID 
		);
		$hookname = $this->getPluginTextDomain();
		wp_clear_scheduled_hook( $hookname, $args );
		
		$ruleset[ 'Status' ] = sprintf( esc_attr__( 'disabled', $this->getPluginTextDomain() ) );
		$this->addUpdateSubIndex( 'Rulesets', $ruleset, $save );
		$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s has been disabled.', $this->getPluginTextDomain() ), $rulesetID ) );
	}
	
	/**
	 * Convert Frequency format into seconds
	 * @param $frequency string the format such as "1d", "2h", "30m", "1d10h30m" etc.
	 * Can also handle time ranges such as "10h-20h", etc.
	 * Can also handle an initial delay time such as "5h+10h", etc.
	 */
	protected function convertTime( $frequency, $returnDelay = false ) {

		if ( empty( $frequency ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Event Frequency is empty.', $this->getPluginTextDomain() ) ) );
			return 0;
		}

		// if more than one Frequency, choose 1 at random
		$matches = explode( ',', $frequency );
		shuffle( $matches );
		$frequency = $matches[0];
		$frequency = trim( $frequency );
		
		// first check for delay time, if any
		$matches = explode( '+', $frequency );
		if ( isset( $matches[ 1 ] ) ) {
			// note: delay does not support 'hourly', 'daily', etc.
			if ( true === $returnDelay )
				return $this->convertTime( $matches[ 0 ] );
			else
				return $this->convertTime( $matches[ 1 ] );
		} else {
			if ( true === $returnDelay )
				return 0;
		}
		
		// first check for time range, if any
		$matches = explode( '-', $frequency );
		if ( isset( $matches[ 1 ] ) ) {
			// note: range does not support 'hourly', 'daily', etc.
			$begin = $this->convertTime( $matches[ 0 ] );
			$end   = $this->convertTime( $matches[ 1 ] );
			if ( $begin <= $end )
				$secs = mt_rand( $begin, $end );
			else
				$secs = mt_rand( $end, $begin );
		} else {
			// recognize standard WordPress values
			if ( $frequency === 'daily' )
				return DAY_IN_SECONDS;
			else if ( $frequency === 'twicedaily' )
				return DAY_IN_SECONDS / 2;
			else if ( $frequency === 'hourly' )
				return HOUR_IN_SECONDS;
			
			$secs = 0;
			if ( preg_match( "/(\d+)\s*d/", $frequency, $matches ) )
				$secs += $matches[ 1 ] * DAY_IN_SECONDS;
			if ( preg_match( "/(\d+)\s*h/", $frequency, $matches ) )
				$secs += $matches[ 1 ] * HOUR_IN_SECONDS;
			if ( preg_match( "/(\d+)\s*m/", $frequency, $matches ) )
				$secs += $matches[ 1 ] * MINUTE_IN_SECONDS;
			if ( preg_match( "/(\d+)\s*s/", $frequency, $matches ) )
				$secs += $matches[ 1 ];
		}

		if ( ! $secs ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: %s, Unknown time format: "%s". Ignored.', $this->getPluginTextDomain() ), __FUNCTION__, $frequency ) );
		}
		return $secs;
	}
	
	/**
	 * Add enabled Rulesets to cron schedules
	 * If using a Frequency range (4h-6h), the interval at the time of wp_schedule_event will be used
	 * @param $rulesetID number of the ruleset to enable
	 */
	public function addToCronSchedules( $schedules ) {

		$rulesets = $this->getOption( 'Rulesets', array ());
		foreach ( $rulesets as $ruleset ) {
			if ( !empty( $ruleset[ 'Frequency' ] ) && isset( $ruleset[ 'Status' ] ) && false !== stripos( $ruleset[ 'Status' ], 'Enabled' ) ) {
				// add event to cron schedule
				$timeInSecs                     = isset( $ruleset[ 'timeInSecs' ] ) ? $ruleset[ 'timeInSecs' ] : $this->convertTime( $ruleset[ 'Frequency' ] );
				$customRecurrence               = $this->getPluginTextDomain() . '-' . $ruleset[ 'ID' ];
				$schedules[ $customRecurrence ] = array(
					'interval' => $timeInSecs,
					'display' => $customRecurrence 
				);
			}
		}
		return $schedules;
	}
	
	/**
	 * If a post is published, check rulesets to see if any events should be restarted.
	 * transition must be from non-publish to publish
	 * @param $new_status string the new transition status
	 * @param $old_status string the old transition status
	 * @param $post the post object
	 */
	public function restartRulesetEvent( $new_status, $old_status, $post ) {
		$ruleset = $this->getVarID();
		if ( $new_status !== 'publish' 
			|| $old_status === 'publish' 
			|| $ruleset
		) {
			return;
		}

		// check for transient WAP sets when updating a post. Do not restart if so.
		if ( get_transient( $this->getPluginTextDomain() . '-updating-post' ) )
			return;

		$post_type = get_post_type( $post->ID );
		$rulesets = $this->getOption( 'Rulesets', array () );
		foreach ( $rulesets as $ruleset ) {
			$post_types = isset( $ruleset[ 'RestartOnPublish' ] ) ? $ruleset[ 'RestartOnPublish' ] : null;
			if ( isset( $ruleset[ 'Status' ] ) 
				&& false !== stripos( $ruleset[ 'Status' ], 'Enabled' ) 
				&& isset( $ruleset[ 'RestartOnPublish' ] ) 
			) {
				// option was a string up to v1.5
				if ( is_string( $ruleset[ 'RestartOnPublish' ] ) && $ruleset[ 'RestartOnPublish' ] !== 'on' ) {
					continue;
				}
				// option is an array after v1.5
				else if ( is_array( $ruleset[ 'RestartOnPublish' ] ) && ! in_array( $post_type, $ruleset[ 'RestartOnPublish' ] ) ) {
					continue;
				}
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, post %d old_status: %s, new_status: %s. Ruleset %s restarting.', $this->getPluginTextDomain() ), __FUNCTION__, $post->ID, $old_status, $new_status, $ruleset[ 'ID' ] ) );
				$this->enableRuleset( $ruleset[ 'ID' ], true );

			}
		}
	}
	
	/**
	 * Ruleset triggered so check all its rules and process
	 * @param $rulesetID number of the ruleset triggered
	 */
	public function processRuleset( $rulesetID, $runOnce = false ) {
		$this->verifyCode();
		$support = $this->getOption( 'Support' );
		if ( empty( $support ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Please enter your License information to Enable Rulesets.', $this->getPluginTextDomain() ) ) );
		}

		set_time_limit( 0 );	// extend script timeout if allowed

		$pid = function_exists( 'getmypid' ) ? getmypid() : '[UNKNOWN]';
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s triggered by PID %d. Checking conditions: %s', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, $pid, print_r( $this->getRuleset( $rulesetID ), true ) ) );

		if ( !$runOnce && !$this->verifySingleCronEvent( $rulesetID ) )
			return;
		if ( ! defined( 'ABSPATH' ) )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: ABSPATH is not defined. Check wp-config.php', $this->getPluginTextDomain() ) ) );

		// first see if eventFrequency is ranged (i.e. 4h-6h) or has been changed, 
		// and we have to update the cron schedule time for next trigger.
		$ruleset = $this->getRuleset( $rulesetID );

		if ( $ruleset[ 'Status' ] === 'disabled' ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset event was called but Ruleset is disabled. Possible issue with your object cache. Aborting.', $this->getPluginTextDomain() ) ) );
		}

		$timeInSecs = $this->convertTime( $ruleset[ 'Frequency' ] );
		$timeInSecs = apply_filters( $this->getPluginTextDomain() . '-epoch-time', $timeInSecs, $ruleset );
		if ( ! isset( $ruleset['timeInSecs'] ) || $timeInSecs !== $ruleset['timeInSecs'] ) {
			$ruleset[ 'timeInSecs' ] = $timeInSecs;
			$this->addUpdateSubIndex( 'Rulesets', $ruleset, true );
		}
		
		// check for Inclusions
		$inclusions = $this->getRulesetOption( $rulesetID, 'Inclusions', null );
		if ( false === $this->passedDateRules( $inclusions, 'inclusions' ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'DEBUG - Ruleset #%s, inclusion dateRules [%s] failed.', $this->getPluginTextDomain() ), $rulesetID, $inclusions ) );
		} 
		else if ( !empty( $inclusions ) )
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - Ruleset #%s, inclusion dateRules [%s] passed.', $this->getPluginTextDomain() ), $rulesetID, $inclusions ) );
		
		// check for Exclusions
		$exclusions = $this->getRulesetOption( $rulesetID, 'Exclusions', null );
		if ( false === $this->passedDateRules( $exclusions, 'exclusions' ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'DEBUG - Ruleset #%s, exclusion dateRules [%s] failed.', $this->getPluginTextDomain() ), $rulesetID, $exclusions ) );
		} 
		else if ( !empty( $exclusions ) )
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - Ruleset #%s, exclusion dateRules [%s] passed.', $this->getPluginTextDomain() ), $rulesetID, $exclusions ) );
		
		// check Max Posts per Day first
		$max_per_day = $this->getRulesetOption( $rulesetID, 'MaxDailyPosts', null );
		$posts_today = 0;
		$today       = date( 'd', current_time( "timestamp" ) );
		if ( true === $this->reachedMaxPerDay( $rulesetID, $max_per_day, $posts_today, $today ) )
			return;
		
		$eligible_posts = $this->queryRulesetPosts( $ruleset );
		
		// post actions:
		$num = 0;
		foreach ( $eligible_posts as $post ) {
			$num++;
			$this->updatePost( $post, $rulesetID, $max_per_day, $posts_today, $today, 'None', $num );
		}
		
		// post recycling:
		$recycleMode   = $this->getRulesetOption( $rulesetID, 'RecycleMode', 'None' );
		$max_per_query = $this->getRulesetOption( $rulesetID, 'MaxNumberPosts', 1 );
		$count         = count( $eligible_posts );
		$num_to_recycle = $max_per_query - $count;
		if ( $recycleMode != 'None' && ( !$max_per_day || $posts_today < $max_per_day ) && $num_to_recycle ) {
			// now recycle any eligible posts found
			$recycle_posts = $this->queryRulesetPosts( $ruleset, $recycleMode );
			$method = $this->getRulesetOption( $rulesetID, 'RecycleMethod', 'Republish' );
			$num = 0;
			foreach ( $recycle_posts as $post ) {
				$num++;
				if ( $method === 'Clone' || $method === 'Replace' ) {
					$new_id = $this->clonePost( $post, $rulesetID );
					$new_post = $new_id ? get_post( $new_id ) : null;
					$this->recyclePost( $new_post, $rulesetID, $max_per_day, $posts_today, $today, $recycleMode, $num );
					if ( $method === 'Replace' && $new_post ) {
						wp_trash_post( $post->ID );
						$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, %s post id %s sent to trash. Title: %s.', $this->getPluginTextDomain() ), $rulesetID, $post->post_status, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $post->post_title ) );
					}
				}
				else {
					$this->recyclePost( $post, $rulesetID, $max_per_day, $posts_today, $today, $recycleMode, $num );
				}
			}
		}
	}
	
	/**
	 * Ruleset triggered so check all its rules and process
	 * @param $rulesetID number of the ruleset triggered
	 */
	protected function importRuleset() {

		$file = $_FILES[ 'wap_import_file' ];
		$filename = $file[ 'name' ];

		if ( empty ( $filename ) )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: %s, No file chosen.', $this->getPluginTextDomain() ), __FUNCTION__ ) );

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, file is: %s', $this->getPluginTextDomain() ), __FUNCTION__, $filename ) );

		if ( $file[ "type" ] != "application/json" )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: %s, File type detected: "%s" but "application/json" required.', $this->getPluginTextDomain() ), __FUNCTION__, $file[ "type" ] ) );

		if ( $file[ "error" ] > 0 )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: %s, Code: "%s". "application/json" exptected.', $this->getPluginTextDomain() ), __FUNCTION__, $file[ "error" ] ) );

		$data = file( $file[ "tmp_name" ] );
		$importRuleset = (array) json_decode( $data[0] );

		if ( empty( $importRuleset ) )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: %s, No data found. FILE: %s.', $this->getPluginTextDomain() ), __FUNCTION__, print_r($file, true) ) );

		// Set Defaults
		$importRuleset[ 'ID' ]  = $this->options[ 'nextID' ] = $this->getOption( 'nextID', 1 );
		$importRuleset[ 'Status' ]  = 'disabled';

		$this->addUpdateSubIndex( 'Rulesets', $importRuleset );
		$this->options[ 'nextID' ]++;
		update_option( $this->getClassName(), $this->options );
		do_action( $this->getPluginTextDomain() . '-saved' );

		$this->addLogMessage( sprintf( esc_html__( 'Import Successful. Ruleset data: %s', $this->getPluginTextDomain() ), print_r( $importRuleset, true ) ) );

	}

	/**
	 * Ruleset triggered so check all its rules and process
	 * @param $rulesetID number of the ruleset triggered
	 */
	protected function exportRuleset( $rulesetID ) {
		$filename = "wap-ruleset-" . $rulesetID . "-" . date_i18n('Y-m-d') . '.json';

		header( "Content-Description: File Transfer" );
		header( "Content-Disposition: attachment; filename={$filename}" );
		header( "Content-Type: application/json; charset=utf-8" );

		$ruleset = $this->getRuleset( $rulesetID );
		echo json_encode( $ruleset ) . PHP_EOL;

		die();
	}

	/**
	 * Simple spintax for the given string like "{Check|Watch} this {awesome|great|amazing} video"
	 */
	private function spintax( $s ) {
		preg_match( '#\{(.+?)\}#uis', $s, $m );
		if ( empty( $m ) ) {
			return $s;
		}
		$t = $m[1];
		if ( strpos( $t, '{' ) !== FALSE ) {
			$t = mb_substr ($t, mb_strrpos($t,'{') + 1 );
		}
		$parts = explode( "|", $t );
		$s = preg_replace( "+\{".preg_quote($t)."\}+is", $parts[array_rand($parts)], $s, 1 );
		return $this->spintax( $s );
	}

	/**
	 * find and replace phrases 
	 */
	private function findReplace( $the_rules, $the_data, $for = 'unknown' ) {
		if ( empty( $the_data ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, %s was empty! Skipping.', $this->getPluginTextDomain() ), __FUNCTION__, $for ) );
			return $the_data;
		}

		// re pattern for REMOVE_SENTENCE.
			$re = '/# Split sentences on whitespace between them.
    				(?<=                # Begin positive lookbehind.
      				[.!?]               # Either an end of sentence punct
    				| [.!?][\'"]        # or end of sentence punct and quote.
				| \n                # or a line break,
    				)                   # End positive lookbehind.
    				(?<!                # Begin negative lookbehind.
      				Mr\.                # Skip either "Mr."
    				| Mrs\.             # or "Mrs.",
    				| Ms\.              # or "Ms.",
    				| Jr\.              # or "Jr.",
    				| Dr\.              # or "Dr.",
    				| Prof\.            # or "Prof.",
    				| Sr\.              # or "Sr.",
				| \s\w\.            # an initial like in "First M. Last"
				| \s\w\.\w\.        # double initial like in "F.M. Last"
				| \s\w.\w\.\w\.     # triple initial like in "F.M.L."
    				)                   # End negative lookbehind.
    				(\s+)               # Split on whitespace between sentences.
    				/mix';

		$rules = explode( PHP_EOL, $the_rules );
		foreach ( $rules as $rule ) {
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, Processing rule: %s.', $this->getPluginTextDomain() ), __FUNCTION__, $for, $rule ) );
			// check for blank line and missing separator
			if ( empty( $rule ) || false === strpos( $rule, '=>' ) ) {
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, rule was empty or missing [=>] separator!', $this->getPluginTextDomain() ), __FUNCTION__, $for ) );
				continue;
			}

			$spinrule = explode( '=>', $rule );
			$regex = '~(?![^<]*>)\b' . preg_quote( trim( $spinrule[0] ) ) . '\b~iu';
			$search = trim( $spinrule[0] );
			$replace =  trim( $spinrule[1] );

			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, searching for pattern [%s] to replace with [%s].', $this->getPluginTextDomain() ), __FUNCTION__, $for, htmlentities( $search) , $replace ) );

			// any special command>?
			if ( $replace === 'REMOVE_SENTENCE' ) {
				$sentences = preg_split( $re, $the_data, -1, PREG_SPLIT_DELIM_CAPTURE );
				for ( $i=0; $i < count( $sentences ); ++$i ) {
					if ( preg_match( $regex, $sentences[$i] ) ) {
						$sentences[$i] = '';
					}
				}
				$result = implode( $sentences );
			}
			else if ( false !== strpos( $replace, 'DATE_FORMAT' ) ) {
				if ( preg_match( '%\(([^\)]+)\)%', $replace, $matches ) ) {
					$format = $matches[1];
				}
				else {
					$format = get_option( 'date_format' );
				}
				$date = date_i18n( $format );
				$result = str_replace( $search, $date, $the_data );
			}
			else if ( false !== strpos( $replace, 'TIME_FORMAT' ) ) {
				if ( preg_match( '%\(([^\)]+)\)%', $replace, $matches ) ) {
					$format = $matches[1];
				}
				else {
					$format = get_option( 'time_format' );
				}
				$time = date_i18n( $format );
				$result = str_replace( $search, $time, $the_data );
			}
			else {

				// is there spintax? figure out spintax replacement ahead of time
				$spintax = $this->spintax( $replace );
				if ( $spintax != $replace ) {
					$replace = trim( $spintax );
				}

				// custom function to keep cases after replacing.
				$result = preg_replace_callback( $regex, function( $matches ) use ( $replace ) {
					$i=0;
					return implode('', array_map( function($char) use ($matches, &$i ) {
						return isset( $matches[0][$i] ) && ctype_upper( $matches[0][$i++] ) ? strtoupper( $char ) : strtolower( $char );
					}, str_split( $replace )));
				}, $the_data );

			} // end else

			// any changes made by this Find and Replace rule?
			if ( ! empty ( $result ) && $result != $the_data ) {
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, rule triggered: %s replaced by %s. Returning [%s]', $this->getPluginTextDomain() ), __FUNCTION__, $for, $spinrule[0], $replace, htmlentities( $result) ) );
				$the_data = $result;
			}
			else {
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, no match found.', $this->getPluginTextDomain() ), __FUNCTION__, $for ) );
			}
		}
		return trim( $the_data );
	}

	private function emailNotification( $postID, $rulesetID, $data ) {
		$post_author_id = get_post_field( 'post_author', $postID );
		$email = get_the_author_meta( 'user_email', $post_author_id ); 
		if ( is_email( $email ) ) {
			$author_name = get_the_author_meta( 'display_name', $post_author_id );
			$from = get_bloginfo( 'admin_email' );
			$subject = sprintf( esc_html__( 'WP Auto Poster made changes to your post on %s.', $this->getPluginTextDomain() ), get_bloginfo( 'name' ) );
			$headers = 'From: ' . $from . "\r\n";
			$headers .= 'Reply-To: ' . $from . "\r\n";
			$permalink = get_permalink( $postID );
			$message = sprintf( esc_html__( 'Post: %s', $this->getPluginTextDomain() ), $permalink );
			$message .= "\r\n" . sprintf( esc_html__( 'Author: %s', $this->getPluginTextDomain() ), $author_name );
			$message .= "\r\n\r\n" . $data;
			require_once( ABSPATH . WPINC . '/pluggable.php' );
			if ( wp_mail( $email, $subject, $message, $headers ) ) 
				$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, Email sent to post author %s about post changes.', $this->getPluginTextDomain() ), $rulesetID, $email ) );
			else
				$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s, Unable to send email to post author %s about post changes.', $this->getPluginTextDomain() ), $rulesetID, $email ) );
		}
		else
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s, post author email %s not valid.', $this->getPluginTextDomain() ), $rulesetID, $email ) );
	}

	private function get_users() {
		$args = array(
				'capability' => array( 'edit_posts' ),
				'fields' => array(
					'ID',
					'display_name' 
				),
				'number' => 1000
			);
		$args = apply_filters( $this->getPluginTextDomain() . '-user-query', $args );

		// Capability queries were only introduced in WP 5.9.
		if ( version_compare( $GLOBALS['wp_version'], '5.9-alpha', '<' ) ) {
    			$args['who'] = 'authors';
    			unset( $args['capability'] );
		}

		$user_query = new \WP_User_Query( $args );
		$users = $user_query->get_results();

		$args = apply_filters( $this->getPluginTextDomain() . '-user-query-results', $users );
		return $users;
	}

	/**
	 * If not in WAP and a post is published, check to see if Ruleset code should be triggered
	 */
	public function checkTransitionTrigger( $new_status, $old_status, $post ) {

		if ( $new_status !== 'publish' )
			return;

		// check for transient WAP sets when updating a post. Do not check for trigger if so
		if ( get_transient( $this->getPluginTextDomain() . '-updating-post' ) )
			return;

		$rulesets = $this->getOption( 'Rulesets', array ());

		foreach ( $rulesets as $ruleset ) {
			if ( isset( $ruleset[ 'Status' ] ) && false !== stripos( $ruleset[ 'Status' ], 'Enabled' ) ) {
				$trigger = $this->getRulesetOption( $ruleset['ID'], 'TriggerOnTransition', 'off' );
				if ( 'on' === $trigger ) {
					$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, Trigger on Transition activated, seeing if post matches.', $this->getPluginTextDomain() ), $ruleset['ID'] ) );
					// see if triggered post matches Post Condition checks:
					$ruleset['LimitPostStatuses'] = 'publish';
					$ruleset['LimitPostIDs'] = $post->ID;
					$eligible_posts = $this->queryRulesetPosts( $ruleset );
					foreach ( $eligible_posts as $e ) {
						if ( $e->ID == $post->ID )	// match!
							$this->processRuleset( $ruleset['ID'] );
					}
				}
			}
		}
	}

	/**
	 * all featured image checks
	 */
	private function featuredImageChecks( $rulesetID, $post ) {
		$modify_featured = $this->getRulesetOption( $rulesetID, 'ModifyFeaturedImage', 'No' );
		$find_featured = $this->getRulesetOption( $rulesetID, 'FindFeaturedImage', 'No' );
		$create_image = $this->getRulesetOption( $rulesetID, 'CreateImageFromTitle', 'No' );
		$set_image = $this->getRulesetOption( $rulesetID, 'FeaturedImageBGImage', "");
		$condition = $this->getRulesetOption( $rulesetID, 'FeaturedImageBGImageCondition', "");

		// do we need to rename the featured image?
		if ( $modify_featured == 'Rename'
			&& $create_image !== 'Yes' 	// no need to rename if going to create anyway...
			&& has_post_thumbnail( $post->ID )
		) {
			$this->renameFeaturedImage( $rulesetID, $post );
		}
		else if ( ( $modify_featured == 'Delete'
			|| $modify_featured == 'Remove' )
			&& has_post_thumbnail( $post->ID )
		) {
			$this->deleteFeaturedImage( $rulesetID, $post, $modify_featured );
		}


		// do we need to check for images to make featured?
		if ( $find_featured == 'Yes'
			&& $create_image !== 'Yes' 	// no need to find if going to create anyway...
			|| ( $find_featured == 'Only if missing' && ! has_post_thumbnail( $post->ID ) )
		) {
			$this->findFeaturedImage( $rulesetID, $post );
		}

		// do we need to set a default featured image?
		if ( $set_image 
			&& $condition == 'Always'
			|| ( $condition == 'Only if missing' && ! has_post_thumbnail( $post->ID ) )
		) {
			$this->setFeaturedImage( $rulesetID, $post );
		}

		// do we need to create a featured image from post title?
		if ( $create_image == 'Yes'
			|| ( $create_image == 'Only if missing' && ! has_post_thumbnail( $post->ID ) )
		) {
			$this->createImageFromText( $rulesetID, $post );
		}
	}

	/**
	 * All common checks and balances when updating a post
	 */
	private function updatePost( $post, $rulesetID, $max_per_day, &$posts_today, $today, $recycleMode = 'None', $num = 0 ) {
		// check Max Posts per Day, again. Ignore if post was just recycled.
		if ( $recycleMode === 'None' 
			&& ! $max_per_day
			&& true === $this->reachedMaxPerDay( $rulesetID, $max_per_day, $posts_today, $today ) 
		) {
			return $this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s max per day reached.', $this->getPluginTextDomain() ), __FUNCTION__ ) );
		}

		$changed = array();
		
		// Post Action: clone?
		$clonePost = $this->getRulesetOption( $rulesetID, 'ClonePost', false );
		if ( true == $clonePost ) {
			$new_post_id = $this->clonePost( $post, $rulesetID );
			if ( ! $new_post_id ) {
				return null;
			}
			$old_post_id = $post->ID;
			$post = get_post( $new_post_id );
		}


		// set short transient so restartOnPublish doesn't trigger.
		$time = apply_filters( $this->getPluginTextDomain() . '-update-post-transient', MINUTE_IN_SECONDS );
		set_transient( $this->getPluginTextDomain() . '-updating-post', 'updating post', $time );

		$args = array(
			'ID' => $post->ID 
		);

		// Post Action: set sticky
		$setPostSticky = $this->getRulesetOption( $rulesetID, 'SetPostSticky', 0 );
		if ( $setPostSticky === 'Stick' ) {
			stick_post( $post->ID );
			$changed[] = esc_html__( 'stickied', $this->getPluginTextDomain() );
		} else if ( $setPostSticky === 'Unstick' ) {
			unstick_post( $post->ID );
			$changed[] = esc_html__( 'unstickied', $this->getPluginTextDomain() );
		}
		
		// Post Action: set post type
		$setPostType = $this->getRulesetOption( $rulesetID, 'SetPostType', 0 );
		if ( $setPostType ) {
			$args[ 'post_type' ] = $setPostType;
			$changed[]           = esc_html__( 'post type set to ', $this->getPluginTextDomain() ) . strtoupper( $setPostType );
		}
		
		// Post Action: set post status
		$setPostStatus = $this->getRulesetOption( $rulesetID, 'SetPostStatus', 0 );
		if ( $setPostStatus && $setPostStatus !== 'inherit' ) {
			$changed[]             = esc_html__( 'post status set to ', $this->getPluginTextDomain() ) . strtoupper( $setPostStatus );
			$args[ 'post_status' ] = $setPostStatus;
		}

		// Post Action: set post date
		$setPostDate = $this->getRulesetOption( $rulesetID, 'SetPostDate', 0 );
		if ( $setPostDate ) {
			if ( $setPostDate === 'Keep Time' ) {
				$args[ 'post_date' ] = $post->post_date;
				$args[ 'post_date_gmt' ] = $post->post_date_gmt;
				$changed[]             = esc_html__( 'post date kept at: ', $this->getPluginTextDomain() ) . $post->post_date;
			}
			else if ( $setPostDate === 'Current Time' ) {
				$args[ 'post_date' ] = current_time( "mysql" );
				$args[ 'post_date_gmt' ] = get_gmt_from_date( $args[ 'post_date' ] );
				$changed[]             = esc_html__( 'post date set to ', $this->getPluginTextDomain() ) . $args[ 'post_date' ];
				$changed[]             = esc_html__( 'post date gmt set to ', $this->getPluginTextDomain() ) . $args[ 'post_date_gmt' ];
			}
		}
		
		// Post Action: set post author
		$setPostAuthor = $this->getRulesetOption( $rulesetID, 'SetPostAuthor', 0 );
		if ( $setPostAuthor == '99999' ) {
			$users = $this->get_users();
			shuffle( $users );
			$setPostAuthor = $users[0]->ID;
		}
		if ( $setPostAuthor ) {
			$args[ 'post_author' ] = $setPostAuthor;
			$changed[]             = esc_html__( 'post author set to ', $this->getPluginTextDomain() ) . strtoupper( get_the_author_meta( 'display_name', $setPostAuthor ) ) . " (user_id {$setPostAuthor})";
		}

		// Post Action: find and replace
		$findreplace = $this->getRulesetOption( $rulesetID, 'FindReplace', array() );
		if ( ! empty( $findreplace ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Find/Replace rules detected: %s', $this->getPluginTextDomain() ), __FUNCTION__, htmlentities( print_r( $findreplace, true ) ) ) );
			$customfindreplace = $this->getRulesetOption( $rulesetID, 'CustomFindReplace', array() );
			$fr_title = $fr_content = $fr_excerpt = false;
			$collapsed = $this->collapseCheckboxes( $customfindreplace );
			foreach ( $collapsed as $key => $value ) {
				if ( $key === 0 && $value === 'on' ) {
					$fr_title = true;
				}
				else if ( $key === 1 && $value === 'on' ) {
					$fr_content = true;
				}
				else if ( $key === 2 && $value === 'on' ) {
					$fr_excerpt = true;
				}
			}
			if ( $fr_title === false	// if nothing is limited, then all are true
					&& $fr_content === false
					&& $fr_excerpt === false ) {
				$fr_title = $fr_content = $fr_excerpt = true;
			}

			if ( $fr_title === true ) {
				$new_title = $this->findReplace( $findreplace, $post->post_title, 'title' );
				if ( $new_title != $post->post_title ) {
					$args[ 'post_title' ] = $new_title;
					$changed[]             = esc_html__( 'find/replace: post title set to ', $this->getPluginTextDomain() ) . '[' . $new_title . ']';
					$args[ 'post_name' ] = ''; // WP will make new slug
				}
			}
			if ( $fr_content === true ) {
				$new_content = $this->findReplace( $findreplace, $post->post_content, 'content' );
				if ( $new_content != $post->post_content ) {
					$args[ 'post_content' ] = $new_content;
					$changed[]             = esc_html__( 'find/replace: post content set to ', $this->getPluginTextDomain() ) . '[' . $new_content . ']';
				}
			}
			if ( $fr_excerpt === true ) {
				$new_excerpt = $this->findReplace( $findreplace, $post->post_excerpt, 'excerpt' );
				if ( $new_excerpt != $post->post_excerpt ) {
					$args[ 'post_excerpt' ] = $new_excerpt;
					$changed[]             = esc_html__( 'find/replace: post excerpt set to ', $this->getPluginTextDomain() ) . '[' . $new_excerpt . ']';
				}
			}
		}
		
		$args = apply_filters( $this->getPluginTextDomain() . '-update-post', $args, $post, $rulesetID );
		
		if ( empty( $changed ) ) {
			if ( $recycleMode === 'None' ) {
				$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, %s post id %s, title: %s - no change', $this->getPluginTextDomain() ), $rulesetID, $post->post_status, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $post->post_title ) );
			}
		}
		else {

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, query result %d, args for wp_update_post are: %s', $this->getPluginTextDomain() ), __FUNCTION__, $num, htmlentities( print_r( $args, true ) ) ) );
		
		// use kses functions so WordPress doesn't strip out suspicious code like iframes
		kses_remove_filters();
		$result = wp_update_post( $args, true ); 
		kses_init_filters();

		if ( is_wp_error( $result ) )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error - Ruleset #%s, wp_update_post returned %s.', $this->getPluginTextDomain() ), $rulesetID, $result->get_error_message() ) );
		else {
			$what_changed = implode( ', ', $changed );
			$sched_time = $this->getRulesetEventTime( $rulesetID, false );
			if ( $sched_time ) {
				$what_changed .= " - " . $sched_time;
			}
			$permalink = get_permalink( $post->ID );
			$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, %s post id %s%s, title: %s - %s', $this->getPluginTextDomain() ), 
				$rulesetID, 
				$post->post_status, 
				"<a target='_blank' href='" . $permalink . "'>" . $post->ID . "</a>", 
				true == $clonePost ? ' (cloned from ' . $old_post_id . ')' : '',
				$post->post_title, 
				$what_changed 
			) );
		}

		$notify_author = $this->getRulesetOption( $rulesetID, 'PostActionNotify', '' );
		if ( $notify_author === 'on' )
			$this->emailNotification( $post->ID, $rulesetID, $what_changed );

		// don't check max posts per day if post was recycled.
		if ( $recycleMode === 'None' && $max_per_day ) {
			$posts_today++;
			set_transient( $this->getPluginTextDomain() . '-posts-today-ruleset-' . $rulesetID, "{$today},{$posts_today}", DAY_IN_SECONDS );
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - Ruleset #%s, posts per day: %s out of %s max', $this->getPluginTextDomain() ), $rulesetID, $posts_today, $max_per_day ) );
		}

		} // end if else things changed
		
		$this->featuredImageChecks( $rulesetID, $post );

// $this->tweetPik(0); // JPH testing....

		// now update any taxonomies on post -------------------------------
		$useTaxonomies = $this->getOption( 'UseTaxonomies', array ());
		foreach ( $useTaxonomies as $key => $useTaxonomy ) {
			$action = $this->getRulesetOption( $rulesetID, 'Action-' . $useTaxonomy, null );
			$terms  = $this->getRulesetOption( $rulesetID, 'Change-' . $useTaxonomy, null );
			$terms  = apply_filters( $this->getPluginTextDomain() . '-set-post-terms', $terms, $post, $rulesetID );
			
			if ( empty( $terms ) )
				continue;
			
			$terms     = array_map( 'intval', $terms ); // make sure ints or bad things happen
			$termnames = array();
			foreach ( $terms as $term ) { // get term names for human readable log msg
				$obj = get_term_by( 'id', $term, $useTaxonomy );
				if ( isset( $obj->name ) )
					$termnames[] = $obj->name;
			}
			
			if ( $action === 'Add' ) {
				$result = wp_set_object_terms( $post->ID, $terms, $useTaxonomy, true );
				if ( is_wp_error( $result ) )
					$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s, post id %s wp_set_object_terms ( %s => %s ) returned %s.', $this->getPluginTextDomain() ), $rulesetID, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $useTaxonomy, print_r( $termnames, true ), $result->get_error_message() ) );
				else
					$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, post id %s taxonomy terms were added ( %s => %s ).', $this->getPluginTextDomain() ), $rulesetID, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $useTaxonomy, print_r( $termnames, true ) ) );
			} else if ( $action === 'Delete' ) {
				$result = wp_remove_object_terms( $post->ID, $terms, $useTaxonomy );
				if ( is_wp_error( $result ) )
					$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s, post id %s wp_remove_object_terms ( %s => %s ) returned %s.', $this->getPluginTextDomain() ), $rulesetID, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $useTaxonomy, print_r( $termnames, true ), $result->get_error_message() ) );
				else
					$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, post id %s taxonomy terms were deleted ( %s => %s ).', $this->getPluginTextDomain() ), $rulesetID, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $useTaxonomy, print_r( $termnames, true ) ) );
			} else if ( $action === 'Replace' ) {
				$result = wp_set_object_terms( $post->ID, $terms, $useTaxonomy, false );
				if ( is_wp_error( $result ) )
					$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s, post id %s wp_set_object_terms ( %s => %s ) returned %s.', $this->getPluginTextDomain() ), $rulesetID, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $useTaxonomy, print_r( $termnames, true ), $result->get_error_message() ) );
				else
					$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, post id %s taxonomy terms were replaced with ( %s => %s ).', $this->getPluginTextDomain() ), $rulsetID, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $useTaxonomy, print_r( $termnames, true ) ) );
			}
		}

		// now update any custom fields on post -------------------------------
		$maxMetaFields = apply_filters( $this->getPluginTextDomain() . '-metafield-actions', 1, $ruleset );
		for ( $x = 1; $x <= $maxMetaFields; $x++ ) {
			$name = $this->getRulesetOption( $rulesetID, 'MetaFieldActionName' . $x, null );
			$value = $this->getRulesetOption( $rulesetID, 'MetaFieldActionValue' . $x, null );
			if ( empty( $name ) )
				continue;
			add_post_meta( $post->ID, $name, $value, true ) 
				|| update_post_meta( $post->ID, $name, $value );
			$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, post id %s custom meta field [%s] set to [%s].', $this->getPluginTextDomain() ), $rulesetID, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $name, $value ) );
		}

		$sleepDelay = $this->getOption( 'SleepDelay', 1 );
		sleep( $sleepDelay );

		delete_transient( $this->getPluginTextDomain() . '-updating-post' );
		
		do_action( $this->getPluginTextDomain() . '-post-updated', $post->ID, $rulesetID );
	}
	
	/**
	 * Clone a post
	 * adapted from https://kinsta.com/knowledgebase/duplicate-page-post-wordpress/#custom-plugin
	 */
	private function clonePost( $post, $rulesetID ) {
		global $wpdb;

		// args for new post
		$args = array(
 			'comment_status' => $post->comment_status,
			'ping_status'    => $post->ping_status,
			'post_author'    => $post->post_author,
			'post_content'   => $post->post_content,
			'post_excerpt'   => $post->post_excerpt,
			'post_name'      => $post->post_name,
			'post_parent'    => $post->post_parent,
			'post_password'  => $post->post_password,
			'post_status'    => 'draft',
			'post_title'     => $post->post_title,
			'post_type'      => $post->post_type,
			'to_ping'        => $post->to_ping,
			'menu_order'     => $post->menu_order
		);

		$args = apply_filters( $this->getPluginTextDomain() . '-clone-post', $args, $post->ID, $rulesetID );
		if ( empty( $args ) ) 	// user abort
			return null;

		// insert the new post
		$new_post_id = wp_insert_post( $args, true );

		if ( is_wp_error( $new_post_id ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Clone Error - Ruleset #%s, wp_insert_post returned %s.', $this->getPluginTextDomain() ), $rulesetID, $new_post_id->get_error_message() ) );
			return null;
		}

		// add taxonomy terms to the new post
		$taxonomies = get_object_taxonomies( $post->post_type );
		foreach ( $taxonomies as $taxonomy ) {
			$post_terms = wp_get_object_terms( $post->ID, $taxonomy, array( 'fields' => 'slugs' ) );
			wp_set_object_terms( $new_post_id, $post_terms, $taxonomy, false );
		}

		// duplicate all post meta
		$post_meta = get_post_meta( $post->ID );
		if( $post_meta ) {

			foreach ( $post_meta as $meta_key => $meta_values ) {

				if( '_wp_old_slug' == $meta_key ) { // do nothing for this meta key
					continue;
				}

				foreach ( $meta_values as $meta_value ) {
					add_post_meta( $new_post_id, $meta_key, $meta_value );
				}
			}
		}

		// now clone comments
		$args = array(
			'post_id'	=> $post->ID
		);
		$comment_query = new \WP_Comment_Query;
		$comments = $comment_query->query( $args );
		if ( $comments ) {
			foreach ( $comments as $comment ) {
				$data = array(
					'comment_post_ID' => $new_post_id,
					'comment_author' => $comment->comment_author,
					'comment_author_email' => $comment->comment_author_email,
					'comment_author_url' => $comment->comment_author_url,
					'comment_author_IP' => $comment->comment_author_IP,
					'comment_date' => $comment->comment_date,
					'comment_date_gmt' => $comment->comment_date_gmt,
					'comment_content' => $comment->comment_content,
					'comment_karma' => $comment->comment_karma,
					'comment_approved' => $comment->comment_approved,
					'comment_agent' => $comment->comment_agent,
					'comment_type' => $comment->comment_type,
					'comment_parent' => $comment->comment_parent,
					'user_id' => $comment->user_id,
				);
				wp_insert_comment( $data );
			}
		}

		$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, %s post id %s CLONED into post id %s. Title: %s.', $this->getPluginTextDomain() ), $rulesetID, $post->post_status, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", "<a target='_blank' href='" . get_post_permalink( $new_post_id ) . "'>" . $new_post_id . "</a>", $post->post_title ) );

		return $new_post_id;
	}

	/**
	 * All common checks and balances when recycling a post
	 */
	private function recyclePost( $post, $rulesetID, $max_per_day, &$posts_today, $today, $recycleMode, $num ) {
		if ( ! $post ) {
			return;
		}
		// check Max Posts per Day, again
		if ( true === $this->reachedMaxPerDay( $rulesetID, $max_per_day, $posts_today, $today ) )
			return;
		
		$noRecycleHooks = $this->getRulesetOption( $rulesetID, 'NoRecycleHooks', '' );
		$sleepDelay = $this->getOption( 'SleepDelay', 1 );
		
		// recycle mode - save post as draft to trigger publish hooks when published
		if ( $noRecycleHooks !== 'on' ) {
			// save the published post as draft, so publishing it again triggers hooks!
			$args = array(
				'ID' => $post->ID,
				'post_status' => 'draft' 
			);
			$args = apply_filters( $this->getPluginTextDomain() . '-draft-post', $args );
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, query result %d, args for wp_update_post are: %s', $this->getPluginTextDomain() ), __FUNCTION__ . '*', $num, htmlentities( print_r( $args, true ) ) ) );
			kses_remove_filters();
			$result = wp_update_post( $args, true );
			kses_init_filters();
			if ( is_wp_error( $result ) ) {
				$this->addLogMessage( sprintf( esc_html__( 'WAP Recycle Error - Ruleset #%s, wp_update_post* returned %s.', $this->getPluginTextDomain() ), $rulesetID, $result->get_error_message() ) );
				return;
			}
			sleep( $sleepDelay );
		}
		
		$args = array(
			'ID' => $post->ID,
			'post_status' => 'publish',
			'post_date' => current_time( "mysql" ) 
		);
		
		$recycleTimes = get_post_meta( $post->ID, '_wap_recycled', true );
		
		$originalPublish = $this->getRulesetOption( $rulesetID, 'OriginalPublish', 'None' );
		if ( $originalPublish != 'None' && !$recycleTimes ) {
			$text = $this->getRulesetOption( $rulesetID, 'OriginalPublishText', 'None' );
			$args = $this->recycleOriginalPublishNote( $args, $post, $originalPublish, $text );
		}
		
		$args = apply_filters( $this->getPluginTextDomain() . '-recycle-post', $args, $post, $rulesetID );
		
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, query result %d, args for wp_update_post are: %s', $this->getPluginTextDomain() ), __FUNCTION__, $num, htmlentities( print_r( $args, true ) ) ) );
		
		// use kses functions so WordPress doesn't strip out suspicious code like iframes
		kses_remove_filters();
		// Note: wp_publish_post() does not update the publish date! using wp_update_post instead!
		$result = wp_update_post( $args, true );
		kses_init_filters();
		
		if ( is_wp_error( $result ) )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Recycle Error - Ruleset #%s, wp_update_post returned %s.', $this->getPluginTextDomain() ), $rulesetID, $result->get_error_message() ) );
		else {
			sleep( $sleepDelay );
			$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, %s post id %s RECYCLED. Title: %s.', $this->getPluginTextDomain() ), $rulesetID, $post->post_status, "<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", $post->post_title ) );
		}

		$notify_author = $this->getRulesetOption( $rulesetID, 'PostRecycleNotify', '' );
		if ( $notify_author === 'on' ) {
			$what_changed = esc_html__( 'Post has been recycled.', $this->getPluginTextDomain() );
			$this->emailNotification( $post->ID, $rulesetID, $what_changed );
		}

		if ( $max_per_day ) {
			$posts_today++;
			set_transient( $this->getPluginTextDomain() . '-posts-today-ruleset-' . $rulesetID, "{$today},{$posts_today}", DAY_IN_SECONDS );
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - Ruleset #%s, posts per day: %s out of %s max', $this->getPluginTextDomain() ), $rulesetID, $posts_today, $max_per_day ) );
		}

		$recycleTimes = empty( $recycleTimes ) ? 1 : $recycleTimes + 1;
		update_post_meta( $post->ID, '_wap_recycled', $recycleTimes );
			
		do_action( $this->getPluginTextDomain() . '-post-recycled', $post->ID, $rulesetID );

		$force_post_actions = $this->getRulesetOption( $rulesetID, 'PostRecycleForceActions', '' );
		if ( $force_post_actions === 'on' ) {
			$this->updatePost( $post, $rulesetID, $max_per_day, $posts_today, $today, $recycleMode, $num );
		}
	}


	/**
	 * Convert relative Post Age to something WordPress wp_query will recognize
	 * @return array of date_query or null.
	 */
	private function getDateQuery( $date_query, $age, $period, $type ) {
		$dq = true === is_numeric( $age ) ? $age . ' days ago' : $age;
		if ( strtotime( $dq ) === false ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: %s "%s" not recognized by strtotime function. Ignored.', $this->getPluginTextDomain() ), $type, $age ) );
		}
		else {
			$date_query = array_merge( $date_query, array(
				$period => $dq
			) );
		}
		return $date_query;
	}
	
	
	/**
	 * Ruleset triggered to check all rules and process
	 * @param $rulesetID number of the ruleset triggered
	 * @return array of WP_Post Objects found
	 */
	private function queryRulesetPosts( $ruleset, $recycleMode = 'None' ) {
		
		$rulesetID = $ruleset['ID'];

		// set default args for WP_Query
		$args = array(
			'posts_per_page' 		=> $this->getRulesetOption( $ruleset, 'MaxNumberPosts', 1 ),
			'meta_query'			=> array(),
                        'no_found_rows'         	=> true,	// don't need pagination
                        'cache_results'         	=> false,	// don't need posts cached
                        'update_post_meta_cache' 	=> false,	// don't need meta cached
                        'update_post_term_cache' 	=> false,	// don't need terms cached
		);
		
		// note: ignore only works on post type 'post' because only 'post' supports it.
		$ignore_sticky_posts = $this->getRulesetOption( $ruleset, 'IgnoreStickies', 'on' );
		if ( $ignore_sticky_posts === 'on' )
			$args[ 'ignore_sticky_posts' ] = true;
		
		$suppress_filters = $this->getRulesetOption( $ruleset, 'SuppressFilters', 'off' );
		if ( $suppress_filters === 'on' ) {
			$args[ 'suppress_filters' ] = true;
		}
		
		$post_types = $this->getRulesetOption( $ruleset, 'LimitPostTypes', null );
		if ( $post_types )
			$args[ 'post_type' ] = $post_types;
		
		if ( $recycleMode != 'None' )
			$post_statuses = 'publish';
		else
			$post_statuses = $this->getRulesetOption( $ruleset, 'LimitPostStatuses', null );
		
		if ( ! $post_statuses 
			|| is_string( $post_statuses ) && '(recycle only)' == $post_statuses 
			|| is_array( $post_statuses ) && in_array( '(recycle only)', $post_statuses )
		) {
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, (recycle only) selected. Skipping query.', $this->getPluginTextDomain() ), __FUNCTION__ ) );
			return array(); // recycle only = no eligible posts
		}
		$args[ 'post_status' ] = $post_statuses;	
		
		$post_authors = $this->getRulesetOption( $ruleset, 'LimitPostAuthors', null );
		if ( $post_authors )
			$args[ 'author' ] = implode( ',', $post_authors );
		
		$useTaxonomies = $this->getOption( 'UseTaxonomies', array ());
		$tax_queries   = array();
		foreach ( $useTaxonomies as $key => $useTaxonomy ) {
			$limit = $this->getRulesetOption( $ruleset, 'Limit-' . $useTaxonomy, 'IN' );
			$terms = $this->getRulesetOption( $ruleset, $useTaxonomy, null );
			if ( $terms && $useTaxonomy === 'category' ) {
				if ( $limit === 'NOT IN' )
					$args[ 'category__not_in' ] = $terms;
				else
					$args[ 'category__in' ] = $terms;
			}
			else if ( $terms && $useTaxonomy === 'post_tag' ) {
				if ( $limit === 'NOT IN' )
					$args[ 'tag__not_in' ] = $terms;
				else
					$args[ 'tag__in' ] = $terms;
			}
			else if ( $terms )
				$tax_queries[] = array(
					'taxonomy' => $useTaxonomy,
					'field' => 'term_id',
					'terms' => $terms,
					'operator' => $limit
				);
		}
		// NOTE: If 'tax_query' is set for a query, the default post_type becomes 'any'. 
		if ( !empty( $tax_queries ) ) {
			$tax_queries[]       = array(
				'relation' => 'AND' 
			);
			$args[ 'tax_query' ] = $tax_queries;
		}
		
		$post_search = $this->getRulesetOption( $ruleset, 'LimitPostSearch', null );
		if ( $post_search )
			$args[ 's' ] = stripslashes( $post_search );
		
		$post_IDs = $this->getRulesetOption( $ruleset, 'LimitPostIDs', null );
		if ( $post_IDs ) {
			$post_IDs_a = explode( ',',  $post_IDs );
			$post_IDs_a = array_filter( $post_IDs_a, 'is_numeric' );
			$posIDs = array_filter( $post_IDs_a, function( $v ) {
				return $v > 0;
			});
			$negIDs = array_map( 'abs', array_filter( $post_IDs_a, function( $v ) {
				return $v < 0;
			}));
			if ( ! empty( $posIDs ) )
				$args[ 'post__in' ] = $posIDs;
			if ( ! empty( $negIDs ) ) {
				$args[ 'post__not_in' ] = $negIDs;
			}
		}
		
		$featured = $this->getRulesetOption( $ruleset, 'LimitFeaturedImages', null );
		if ( $featured === 'Has' )
			$args[ 'meta_query' ] = array_merge( $args[ 'meta_query'],
				array( array(
					'key'	=> '_thumbnail_id'
				) )
			);
		else if ( $featured === 'Missing' )
			$args[ 'meta_query' ] = array_merge( $args[ 'meta_query'],
				array( array(
					'key'	=> '_thumbnail_id',
					'compare'	=> 'NOT EXISTS'
				) )
			);
		
		$date_query = array();

		if ( $recycleMode != 'None' ) {
			$orderby = $recycleMode;
			$minAge  = $this->getRulesetOption( $ruleset, 'MinRecycleAge', 0 );
			$maxAge  = $this->getRulesetOption( $ruleset, 'MaxRecycleAge', 0 );
			if ( $minAge ) {
				$date_query = $this->getDateQuery( $date_query, $minAge, 'before', 'Minimum Recycle Age' );
			}
			if ( $maxAge ) {
				$date_query = $this->getDateQuery( $date_query, $maxAge, 'after', 'Maximum Recycle Age' );
			}
			$maxRecycleTimes = $this->getRulesetOption( $ruleset, 'MaxRecycleTimes', 0 );
			if ( $maxRecycleTimes ) {
				$args[ 'meta_query' ] = array_merge( $args[ 'meta_query' ],
					array(
					'relation' => 'OR',
					array(
						'key' => '_wap_recycled',
						'value' => $maxRecycleTimes,
						'compare' => '<' 
					),
					array(
						'key' => '_wap_recycled',
						'value' => '',
						'compare' => 'NOT EXISTS' 
					) 
					)
				);
			}
		} else {
			$minAge     = $this->getRulesetOption( $ruleset, 'MinPostAge', 0 );
			$maxAge     = $this->getRulesetOption( $ruleset, 'MaxPostAge', 0 );
			if ( $minAge ) {
				$date_query = $this->getDateQuery( $date_query, $minAge, 'before', 'Minimum Post Age' );
			}
			if ( $maxAge ) {
				$date_query = $this->getDateQuery( $date_query, $maxAge, 'after', 'Maximum Post Age' );
			}
			$orderby = $this->getRulesetOption( $ruleset, 'QueryOrderBy', 'Oldest' );
		}

		$requiredComments	= $this->getRulesetOption( $ruleset, 'RequiredComments', 0 );
		if ( $requiredComments ) {
			$operator = $this->getRulesetOption( $ruleset, 'CommentsOperator', '>=' );
			if ( $operator == 'gt=' )
				$operator = '>=';
			else if ( $operator == 'lt=' )
				$operator = '<=';
			$args[ 'comment_count' ] = array(
				'value'		=> $requiredComments,
				'compare'	=> $operator
			);
		}

		$maxMetaFields = apply_filters( $this->getPluginTextDomain() . '-metafield-cond', 1, $ruleset );
		$defaultRelation = apply_filters( $this->getPluginTextDomain() . '-metafield-cond-relation', 'AND', $ruleset );
		for ( $x = 1; $x <= $maxMetaFields; $x++ ) {

		$metaName = $this->getRulesetOption( $ruleset, 'MetaFieldCondName' . $x, '' );
		$metaCompare = $this->getRulesetOption( $ruleset, 'MetaFieldCondCompare' . $x, '=' );
		$metaValue = $this->getRulesetOption( $ruleset, 'MetaFieldCondValue' . $x, '' );
		if ( $metaName && $metaValue ) {
				$args[ 'meta_query' ] = array_merge( $args[ 'meta_query' ],
					array(
						'relation'	=> $defaultRelation,
						array(
							'key' => $metaName,
							'compare' => $metaCompare,
							'value' => $metaValue
						),
					),
				);
		}
		else if ( $metaName && ! $metaValue ) {
				$args[ 'meta_query' ] = array_merge( $args[ 'meta_query' ],
					array(
						'relation'	=> $defaultRelation,
						array(
							'key' => $metaName,
						),
					),
				);
		}
		else if ( !$metaName && $metaValue ) {
				$args[ 'meta_query' ] = array_merge( $args[ 'meta_query' ],
					array(
						'relation'	=> $defaultRelation,
						array(
							'compare' => $metaCompare,
							'value' => $metaValue
						),
					),
				);
		}
		} // end for x
		
		if ( $orderby === 'CommentCount' ) {
			$args[ 'order' ]   = 'DESC';
			$args[ 'orderby' ] = 'comment_count';
		} else if ( $orderby === 'Oldest' ) {
			$args[ 'order' ]   = 'ASC';
			$args[ 'orderby' ] = 'date';
		} else if ( $orderby === 'Newest' ) {
			$args[ 'order' ]   = 'DESC';
			$args[ 'orderby' ] = 'date';
		} else if ( $orderby === 'Random' ) {
			$args[ 'order' ]   = 'DESC';
			$args[ 'orderby' ] = 'rand';
		} else if ( $orderby === 'SearchRelevance' ) {
			$args[ 'order' ]   = 'DESC';
			$args[ 'orderby' ] = 'relevance';
		}
		// recycle mode only:
		else if ( $orderby === 'Daily' ) {
			$args[ 'order' ] = 'ASC';
			$today           = getdate();
			$date_query      = array_merge( $date_query, array(
				'year' 	=> $today[ 'year' ],
				'month'	=> $today[ 'month' ],
				'day' 	=> $today[ 'day' ]
			) );
		}
		else if ( $orderby === 'Weekly' ) {
			$args[ 'order' ] = 'ASC';
			$today           = getdate();
			$date_query      = array_merge( $date_query, array(
				'dayofweek_iso' => $today[ 'wday' ] 
			) );
			if ( !$minAge ) // make sure not to include today 
				$date_query = array_merge( $date_query, array(
					'before' => '1 days ago' 
				) );
		}
		else if ( $orderby === 'Monthly' ) {
			$args[ 'order' ] = 'ASC';
			$today           = getdate();
			$date_query      = array_merge( $date_query, array(
				'day' => $today[ 'mday' ] 
			) );
			if ( !$minAge ) // make sure not to include today 
				$date_query = array_merge( $date_query, array(
					'before' => '1 days ago' 
				) );
		}
		else if ( $orderby === 'OTD' ) {
			$args[ 'order' ] = 'ASC';
			$today           = getdate();
			$date_query      = array_merge( $date_query, array(
				'month' => $today[ 'mon' ],
				'day' => $today[ 'mday' ] 
			) );
			if ( !$minAge ) // make sure not to include today
				$date_query = array_merge( $date_query, array(
					'before' => '1 days ago' 
				) );
		}
		
		if ( !empty( $date_query ) )
			$args[ 'date_query' ] = $date_query;

		if ( empty( $args['meta_query'] ) ) 
			unset( $args['meta_query'] );

		$customSearch = $this->getRulesetOption( $ruleset, 'CustomSearch', array() );
		$collapsed = $this->collapseCheckboxes( $customSearch );
		foreach ( $collapsed as $key => $value ) {
			if ( $key === 0 && $value === 'on' ) {
				$args[ 'wap_search' ][] = 'title';
			}
			else if ( $key === 1 && $value === 'on' ) {
				$args[ 'wap_search' ][] = 'content';
			}
			else if ( $key === 2 && $value === 'on' ) {
				$args[ 'wap_search' ][] = 'excerpt';
			}
		}

		$args = apply_filters( $this->getPluginTextDomain() . '-query-posts', $args, $rulesetID );
		
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, args for WP_Query are: %s', $this->getPluginTextDomain() ), __FUNCTION__, print_r( $args, true ) ) );
		
		// give user a chance to use own custom code to order post
		$posts = apply_filters( $this->getPluginTextDomain() . '-wp-query-posts', null, $args, $rulesetID );
		if ( ! is_array( $posts ) ) {
			$query = new \WP_Query( $args );
			$posts = isset( $query->posts ) ? $query->posts : array();
		}
		
		if ( $args[ 'posts_per_page' ] <= 1 )
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, results from WP_Query are: %s', $this->getPluginTextDomain() ), __FUNCTION__, htmlentities( print_r( $posts, true ) ) ) );
		else
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, found %s results from WP_Query', $this->getPluginTextDomain() ), __FUNCTION__, $query->post_count ) );
		
		$type = $recycleMode != 'None' ? 'recyclable' : 'eligible';
		if ( empty( $posts ) )
			$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, WP_Query did not return any %s posts.', $this->getPluginTextDomain() ), $rulesetID, $type ) );
		
		return $posts;
	}
	
	/**
	 * Check max per day limit using transient.
	 * @param $rulesetID number of the ruleset to be checked
	 * @param $max_per_day number of the option 'Max Posts Per Day'
	 * @param $posts_today number of posts published by this ruleset today
	 * @param $today number of the day of the month
	 */
	private function reachedMaxPerDay( $rulesetID, $max_per_day, &$posts_today, $today ) {
		
		if ( !$max_per_day ) // no limit
			return false;
		
		$pToday = get_transient( $this->getPluginTextDomain() . '-posts-today-ruleset-' . $rulesetID );
		
		if ( $pToday ) {
			// transient in format 'day of month, number of posts'
			list( $tdate, $posts_today ) = explode( ',', $pToday );
			if ( $today != $tdate )
				$posts_today = 0; // new day, start count over
			if ( $posts_today >= $max_per_day ) {
				$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s, reached %s maximum number of posts for today.', $this->getPluginTextDomain() ), $rulesetID, $max_per_day ) );
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Check if Inclusion/Exclusion date rules pass current date/time
	 * @param $dateRules string a comma delimited entry for specific date/times
	 * @param $inclusion boolean whether this is an Inclusion rule (true) or Exclusion rule (false)
	 * @return boolean if this rule passes or fails
	 */
	private function passedDateRules( $dateRules, $type ) {
		if ( empty( $dateRules ) )
			return true;

		$user_check = apply_filters( $this->getPluginTextDomain() . '-passedDateRules', false, $dateRules, $type);
		if ( $user_check ) {
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, filter %s returned true for %s.', $this->getPluginTextDomain() ), __FUNCTION__, $type, $this->getPluginTextDomain() . '-passedDateRules', $date ) );
			return true;
		}
		
		$checkTime = current_time( "timestamp" );
		
		$dates     = explode( ',', $dateRules );
		
		$passed = true;
		
		foreach ( $dates as $date ) {

			$date = trim( $date );

			if ( empty( $date ) ) // something like "0000-0900,2100-0000," would leave an empty date
				return true;
			
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, checking condition \'%s\'. Check time is %s', $this->getPluginTextDomain() ), __FUNCTION__, $type, $date, date( "D M j G:i:s T Y", $checkTime ) ) );
			
			$yearCheck = $monthCheck = $dayCheck = $MCheck = $DCheck = $hoursCheck = $startHour = $endHour = null;

			// next look for format (something):hhhh-hhhh
			$parts = explode( ':', $date );
			if ( count( $parts ) >= 2 ) {
				$date       = $parts[ 0 ];
				$hoursCheck = $parts[ 1 ];
			}

			// first look for a relative format, like "first friday of the month"
			// also catches "Mon" or "Friday"
			if ( false !== ( $rel = strtotime( $date ) ) ) {	
				// https://www.php.net/manual/en/datetime.formats.relative.php
				$reldate = $date;
				$date = date( "Y-m-d", $rel );
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, converted relative format \'%s\' to \'%s\'', $this->getPluginTextDomain() ), __FUNCTION__, $type, $reldate, $date ) );
			}

			// formats: YYYY-MM-DD and variants
			$parts = explode( '-', $date );
			
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, date parts (YYYY-MM-DD) are %s', $this->getPluginTextDomain() ), __FUNCTION__, $type, print_r( $parts, true ) ) );

			if ( count( $parts ) >= 3 ) {
				// YYYY-MM-DD
				$yearCheck  = $parts[ 0 ];
				$monthCheck = $parts[ 1 ];
				$dayCheck   = $parts[ 2 ];
			} else if ( count( $parts ) === 2 ) {
				if ( strlen( $parts[ 0 ] ) >= 4 && strlen( $parts[ 1 ] >= 4 ) )
					$hoursCheck = $date; // just hhhh-hhhh
				else if ( strlen( $parts[ 0 ] ) >= 4 ) {
					// YYYY-MM
					$yearCheck  = $parts[ 0 ];
					$monthCheck = $parts[ 1 ];
				} else {
					// MM-DD
					$monthCheck = $parts[ 0 ];
					$dayCheck   = $parts[ 1 ];
				}
			} else if ( count( $parts ) <= 1 ) {
				$date = ucwords( $date );
				if ( is_numeric( $date ) && strlen( $date ) >= 4 )
					$yearCheck = $date;
				else if ( is_numeric( $date ) )
					$dayCheck = $date;
				else if ( true === $this->checkDateValidity( $date, "M" ) )
					$MCheck = $date;
				else if ( true === $this->checkDateValidity( $date, "D" ) )
					$DCheck = $date;
				else {
					$this->addLogMessage( sprintf( esc_html__( 'WAP Error: function %s for %s, \'%s\' is not recognized. Ignored.', $this->getPluginTextDomain() ), __FUNCTION__, $type, $date ) );
					continue;
				}
			}
			
			if ( !empty( $hoursCheck ) ) {
				$parts = explode( '-', $hoursCheck );
				$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s for %s, hours parts are %s', $this->getPluginTextDomain() ), __FUNCTION__, $type, print_r( $parts, true ) ) );
				if ( count( $parts ) >= 2 ) {
					$startHour = $parts[ 0 ];
					$endHour   = $parts[ 1 ];
				} else {
					$this->addLogMessage( sprintf( esc_html__( 'WAP Error: function %s for %s, %s is not recognized. Ignored.', $this->getPluginTextDomain() ), __FUNCTION__, $type, $date ) );
					continue;
				}
				$time24 = date( "Hi", $checkTime );
			}

			if ( $yearCheck && !is_numeric( $yearCheck ) || $monthCheck && !is_numeric( $monthCheck ) || $dayCheck && !is_numeric( $dayCheck ) || $startHour && !is_numeric( $startHour ) || $endHour && !is_numeric( $endHour ) ) {
				$this->addLogMessage( sprintf( esc_html__( 'WAP Error: function %s for %s, %s is not recognized. Ignored.', $this->getPluginTextDomain() ), __FUNCTION__, $type, $date ) );
				continue;
			}
			
			if ( 'inclusions' === $type ) {
				if ( $hoursCheck && ( $time24 < $startHour || $time24 > $endHour ) )
					$passed = false;
				else if ( $yearCheck && $yearCheck != date( "Y", $checkTime ) )
					$passed = false;
				else if ( $monthCheck && $monthCheck != date( "m", $checkTime ) )
					$passed = false;
				else if ( $dayCheck && $dayCheck != date( "d", $checkTime ) )
					$passed = false;
				else if ( $MCheck && false === $this->check3Dates( $MCheck, "M" ) )
					$passed = false;
				else if ( $DCheck && false === $this->check3Dates( $DCheck, "D" ) )
					$passed = false;
				else // only one inclusion rule needs to pass for entire check to pass
					return true;
			} else {
				$passed = false;
				if ( $hoursCheck && ( $time24 < $startHour || $time24 > $endHour ) )
					$passed = true;
				if ( $yearCheck && $yearCheck != date( "Y", $checkTime ) )
					$passed = true;
				if ( $monthCheck && $monthCheck != date( "m", $checkTime ) )
					$passed = true;
				if ( $dayCheck && $dayCheck != date( "d", $checkTime ) )
					$passed = true;
				if ( $MCheck && false === $this->check3Dates( $MCheck, "M" ) )
					$passed = true;
				if ( $DCheck && false === $this->check3Dates( $DCheck, "D" ) )
					$passed = true;
				if ( false === $passed ) // only one exclusion rule needs to fail
					return false;
			}
			
		} // end for loop
		return $passed;
	}
	
	/**
	 * Check if 3 letter months/days setting are valid
	 * @param string $check in format Jan/Feb/Mar or Mon/Tue/Wed
	 * @param string $which either "M" or "D"
	 * @return boolean true or false
	 */
	private function checkDateValidity( $check, $which ) {
		$monthsofYear = array(
			'Jan',
			'Feb',
			'Mar',
			'Apr',
			'May',
			'Jun',
			'Jul',
			'Aug',
			'Sep',
			'Oct',
			'Nov',
			'Dec',
			'January',
			'February',
			'March',
			'April',
			'May',
			'June',
			'July',
			'August',
			'September',
			'October',
			'November',
			'December' 
		);
		$daysofWeek   = array(
			'Mon',
			'Tue',
			'Wed',
			'Thu',
			'Fri',
			'Sat',
			'Sun',
			'Monday',
			'Tuesday',
			'Wednesday',
			'Thursday',
			'Friday',
			'Saturday',
			'Sunday' 
		);
		
		$whichCheck = $which === 'M' ? $monthsofYear : $daysofWeek;
		$dates = explode( '/', $check );
		foreach ( $dates as $date ) {
			if ( !in_array( $date, $whichCheck ) ) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Check if current day or month matches format for today
	 * If long form, change format to match. 
	 * @param string $check in format Jan/Feb/Mar or Mon/Tue/Wed
	 * @param string $format the format to check either "M" or "D"
	 * @return boolean true or false
	 */
	private function check3Dates( $check, $format, $checkTime = null ) {

		if ( ! $checkTime )
			$checkTime = current_time( 'timestamp' );

		$dates = explode( '/', $check );
		
		foreach ( $dates as $date ) {
			// will strlen still work on i18n days???
			if ( strlen( $date ) <= 3 ) // Mon, Tue, Jan, Feb, etc.
				$match = date( $format, $checkTime );
			elseif ( $format === 'D' ) // Sunday through Saturday
				$match = date( 'l', $checkTime );
			elseif ( $format === 'M' ) // January through December
				$match = date( 'F', $checkTime );

			if ( $date === $match )
				return true;
		}
		return false;
	}
	
	/**
	 * Recursively sort an array of taxonomy terms hierarchically. Child categories will be
	 * placed under a 'children' member of their parent term.
	 * @param Array   $cats     taxonomy term objects to sort
	 * @param Array   $into     result array to put them in
	 * @param integer $parentId the current parent ID to put them in
	 */
	private function sortTermsHierarchicaly( &$cats, &$into, $parentId = 0 ) {
		foreach ( $cats as $i => $cat ) {
			if ( $cat->parent === $parentId ) {
				$into[ $cat->term_id ] = $cat;
				unset( $cats[ $i ] );
			}
		}
		
		foreach ( $into as $topCat ) {
			$topCat->children = array();
			$this->sortTermsHierarchicaly( $cats, $topCat->children, $topCat->term_id );
		}
	}
	
	/**
	 * Convert taxonomy term objects into id:name array
	 * Indent taxnomies when they are children according to level down
	 * @param array $terms the list of terms for this taxonomy
	 * @param array $into the re-ordered list of terms for this taxonomy, now in id->name format
	 * @param integer $level the depth of the current tree, used for indenting the name
	 * Example: parent term
	 *          - child term
	 *          -- grandchild term
	 */
	private function convertTerms( $terms, &$into, $level = 0 ) {
		foreach ( $terms as $term ) {
			$prefix = '';
			for ( $x = 0; $x < $level; $x++ )
				$prefix .= '&#8212;' . ' ';
			$into[ $term->term_id ] = $prefix . $term->name;
			if ( !empty( $term->children ) )
				$this->convertTerms( $term->children, $into, $level + 1 );
		}
	}
	
	/**
	 * Sometimes WordPress gets confused and schedules additional events. 
	 * Make sure there is only one before executing event trigger
	 * @param integer $rulesetID  identifies the scheduled event(s) by this ID
	 */
	private function verifySingleCronEvent( $rulesetID ) {
		$crons = get_option( 'cron' );
		$args  = array(
			(int) $rulesetID 
		);
		$key   = md5( serialize( $args ) );
		
		$events = 0;
		$hook   = $this->getPluginTextDomain();
		
		foreach ( $crons as $timestamp => $cron ) { // note: timestamp is for future events, not current
			if ( isset( $cron[ $hook ][ $key ] ) ) {
				$events++;
			}
		}
		
		if ( $events > 1 ) { // "There can be only one!" - Highlander
			wp_clear_scheduled_hook( $hook, $args );
			$customRecurrence = $hook . '-' . $rulesetID;
			wp_schedule_event( time() + 10, $customRecurrence, $hook, $args );
			$this->addLogMessage( sprintf( esc_html__( 'WAP Warning - function %s, %s hook had %d events scheduled. Resetting.', $this->getPluginTextDomain() ), __FUNCTION__, $customRecurrence, $events ) );
			return 0;
		}
		
		return 1;
	}
	
	/**
	 * Whether to add the post's original publish date when recycling an old post.
	 * Will only do this once because it checks for the 'wap-recycled' class used in note.
	 * @param array $args the param args for upcoming call to wp_update_post
	 * @param WP_POST $post the current post we are recycling
	 * @param string $originalPublish is the setting for this ruleset
	 * return modified $args
	 */
	public function recycleOriginalPublishNote( $args, $post, $originalPublish, $note ) {
		
		if ( empty( $note ) )
			$note = 'This post was originally published on %%DATE%%';

		$note = str_ireplace( "%%DATE%%", get_the_date( '', $post->ID ), $note );
		$note = str_ireplace( "%%AUTHOR%%", get_the_author_meta( 'display_name', $post->post_author ), $note );
		
		$note = apply_filters( $this->getPluginTextDomain() . '-recycleOriginalPublishNote', $note, $post );
		if ( $originalPublish === 'Prepend' )
			$args[ 'post_content' ] = $note . $post->post_content;
		else
			$args[ 'post_content' ] = $post->post_content . $note;
		
		return $args;
	}

	/* ====================================================================================== */

	/**
	 * Customize WordPress search function. Can limit to any combo of title,content,excerpt,custom field
	 * @param $search
	 * @param $wp_query
	 * return modified $search string
	*/
	public function wapCustomSearch( $search, $wp_query ) {
		global $wpdb;

		if ( empty( $search ) ) {
			return $search; // skip processing - no search term in query
		}

		$q = $wp_query->query_vars;
		if ( empty( $q[ 'wap_search' ] ) ) {
			return $search;
		}

		$n = ! empty( $q[ 'exact' ] ) ? '' : $wpdb->placeholder_escape(); // WordPress escapes "%" since 4.8.3 so we can't use percent character directly.
	
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, original WHERE search: %s', $this->getPluginTextDomain() ), __FUNCTION__, $search ) );

		$search_array = array();
		$keep_s = false;
	
		if ( in_array( 'title', $q[ 'wap_search' ] ) ) {
			$search_title = $searchand = '';
			foreach ( (array)$q[ 'search_terms' ] as $term ) {
				if ( substr( $term, 0, 1 ) === '-' ) {
					$comp = 'NOT LIKE';
					$term = ltrim( $term, '-' );
				}
				else
					$comp = 'LIKE';
				$term = esc_sql( $wpdb->esc_like($term) );
				$search_title .= "{$searchand}($wpdb->posts.post_title {$comp} '{$n}{$term}{$n}')";
				$searchand = ' AND ';
			}
			if ( ! empty( $search_title ) ) {
				$search_array[] = $search_title;
				$keep_s = true;
			}
		}
	
		if ( in_array( 'content', $q[ 'wap_search' ] ) ) {
			$search_content = $searchand = '';
			foreach ( (array)$q[ 'search_terms' ] as $term ) {
				if ( substr( $term, 0, 1 ) === '-' ) {
					$comp = 'NOT LIKE';
					$term = ltrim( $term, '-' );
				}
				else
					$comp = 'LIKE';
				$term = esc_sql( $wpdb->esc_like($term) );	
				$search_content .= "{$searchand}($wpdb->posts.post_content {$comp} '{$n}{$term}{$n}')";
				$searchand = ' AND ';
			}
			if ( ! empty( $search_content ) ) {
				$search_array[] = $search_content;
				$keep_s = true;
			}
		}
	
		if ( in_array( 'excerpt', $q[ 'wap_search' ] ) ) {
			$search_excerpt = $searchand = '';
			foreach ( (array)$q[ 'search_terms' ] as $term ) {
				if ( substr( $term, 0, 1 ) === '-' ) {
					$comp = 'NOT LIKE';
					$term = ltrim( $term, '-' );
				}
				else
					$comp = 'LIKE';
				$term = esc_sql( $wpdb->esc_like($term) );
				$search_excerpt .= "{$searchand}($wpdb->posts.post_excerpt {$comp} '{$n}{$term}{$n}')";
				$searchand = ' AND ';
			}
			if ( ! empty( $search_excerpt ) ) {
				$search_array[] = $search_excerpt;
				$keep_s = true;
			}
    		}
	
		$search = implode( ' OR ', $search_array );
		
		if ( ! empty( $search ) ) {
			$search = " AND ({$search}) ";
			if ( ! is_user_logged_in() )
				$search .= " AND ($wpdb->posts.post_password = '') ";
		}
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, modified WHERE search: %s', $this->getPluginTextDomain() ), __FUNCTION__, $search ) );

		return $search;
	}

	/* ====================================================================================== */

	/**
	 * delete featured image completely
	 * @param $rulesetID
	 * @param $post
	 * return true if successful or false if fail 
	*/
	private function deleteFeaturedImage( $rulesetID, $post, $modify ) {

		$thumbnailID = get_post_thumbnail_id( $post->ID );

		if ( ! $thumbnailID ) return;

		delete_post_thumbnail( $post );

		$force_delete = $modify === 'Delete' ? true : false;
		$tmp = wp_delete_attachment( $thumbnailID, $force_delete );

		if ( is_wp_error( $tmp ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error, %s for post id %d returned %s.', $this->getPluginTextDomain() ), __FUNCTION__, $post->ID, $tmp->get_error_message() ) );
		}

		$this->addLogMessage( sprintf( esc_html__( 'Ruleset #%s %sd the Featured Image %s for post id %s.', $this->getPluginTextDomain() ), $rulesetID, $modify, $thumbnailID, $post->ID ) );
	}

	/* ====================================================================================== */

	/**
	 * rename featured image to match the post title
	 * @param $rulesetID
	 * @param $post
	 * return true if successful or false if fail 
	*/
	private function renameFeaturedImage( $rulesetID, $post ) {

		$thumbnailID = get_post_thumbnail_id( $post->ID );
		$origFileUrl = wp_get_attachment_url( $thumbnailID );

		$result = $this->fetchImageAsAttachment( $origFileUrl, $post->ID, true, $post->post_title );

		if ( ! $result || is_wp_error( $result['attachID'] ) )
			return false;

		$this->addLogMessage( sprintf( esc_html__( 
			'Ruleset #%s, post id %s featured image renamed to [%s] and set to attachment id %s.', $this->getPluginTextDomain() ), 
				$rulesetID, 
				"<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", 
				$result['newFile'],
				"<a target='_blank' href='" . get_post_permalink( $result['attachID'] ) . "'>" . $result['attachID'] . "</a>"
	) );

		// delete old attachment data
		$tmp = wp_delete_attachment( $thumbnailID, true );	

		if ( is_wp_error( $tmp ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error, %s for post id %d returned %s.', $this->getPluginTextDomain() ), __FUNCTION__, $post->ID, $tmp->get_error_message() ) );
		}

		return true;
	}

	/* ====================================================================================== */

	/**
	* adapted from https://wordpress.stackexchange.com/questions/30284/media-sideload-image-file-name
 	* Download an image from the specified URL and attach it to a post, optionally setting it featured.
 	* Additional functionality: ability override actual filename, and to pass $post_data to override values in wp_insert_attachment (original only allowed $desc). Also sets the alt tag automatically.
 	*
 	* @param string $url (required) The URL of the image to download
 	* @param int $post_id (required) The post ID the media is to be associated with
 	* @param bool $thumb (optional) Whether to make this attachment the Featured Image for the post (post_thumbnail)
 	* @param string $filename (optional) Replacement filename for the URL filename (do not include extension)
 	* @param array $post_data (optional) Array of key => values for wp_posts table (ex: 'post_title' => 'foobar', 'post_status' => 'draft')
	* @return false on error or 
 	* @return array with 'attachID': int|object The ID of the attachment or a WP_Error on failure
 	* @return array with 'newFile': the name of the file saved
	*/

	private function fetchImageAsAttachment( $url = null, $post_id = null, $thumb = true, $filename = null, $post_data = array() ) {

		if ( !$url || !$post_id )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error, %s( %s ) missing post ID %s.', $this->getPluginTextDomain() ), __FUNCTION__, $url, $post_id ) );

		// verify source image is greater size than media library dimensions for thumbnail:
		$minW = get_option( 'thumbnail_size_w' );
		$minH = get_option( 'thumbnail_size_h' );
		$imageInfo = wp_getimagesize( $url );
		if ( isset( $imageInfo[0] ) && $imageInfo[0] < $minW ) {
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - Image [%s] width %d < declared thumbnail width %d, skipping this image.',  
				$this->getPluginTextDomain() ), 
				$url,
				$imageInfo[0],
				$minW
			) );
			return false;
		}
		else if ( isset( $imageInfo[1] ) && $imageInfo[1] < $minH ) {
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - Image [%s] height %d < declared thumbnail height %d, skipping this image.',  
				$this->getPluginTextDomain() ), 
				$url,
				$imageInfo[1],
				$minH
			) );
			return false;
		}

		$title = $filename;

		require_once( ABSPATH . 'wp-admin/includes/file.php' );

		// Download file to temp location, returns full server path to temp file
		// ex; /home/user/public_html/mysite/wp-content/26192277_640.tmp
		$tmp = download_url( $url );

		// If error storing temporarily, unlink
		if ( is_wp_error( $tmp ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error, %s( %s ), download_url returned %s.', $this->getPluginTextDomain() ), __FUNCTION__, $url, $tmp->get_error_message() ) );
		}

		preg_match('/[^\?]+\.(jpg|JPG|jpe|JPE|jpeg|JPEG|gif|GIF|png|PNG)/', $url, $matches);    // fix file filename for query strings
		$url_filename = basename($matches[0]);			// extract filename from url for title
		$url_type = wp_check_filetype($url_filename);		// determine file type (ext and mime/type)

		// override filename if given, reconstruct server path
		if ( !empty( $filename ) ) {
			$filename = sanitize_file_name($filename);
			$tmppath = pathinfo( $tmp );			// extract path parts
			$new = $tmppath['dirname'] . "/". $filename . "." . $tmppath['extension'];          // build new path
			if ( ! rename($tmp, $new) ) {				// renames temp file on server
				return $this->addLogMessage( sprintf( esc_html__( 'WAP Error, %s( %s ), rename failed.', $this->getPluginTextDomain() ), __FUNCTION__, $url ) );
			}
       		 	$tmp = $new;				// push new filename (in path) to be used in file array later
		}

		// assemble file data (should be built like $_FILES since wp_handle_sideload() will be using)
		$file_array = array();
		$file_array['tmp_name'] = $tmp;			// full server path to temp file

		if ( !empty( $filename ) ) {
			// user given filename for title, add original URL extension
			$file_array['name'] = $filename . "." . $url_type['ext'];
		} else {
			$file_array['name'] = $url_filename;	// just use original URL filename
		}

		// set additional wp_posts columns
		if ( empty( $post_data['post_title'] ) && $title ) {
			$post_data['post_title'] = $title;
		}
		else if ( empty( $post_data['post_title'] ) ) {
			// just use the original filename (no extension)
			$post_data['post_title'] = basename($url_filename, "." . $url_type['ext']);
		}

		// make sure gets tied to a post parent
		if ( empty( $post_data['post_parent'] ) ) {
			$post_data['post_parent'] = $post_id;
		}

		// required libraries for media_handle_sideload and wp_generate_attachment_metadata
		require_once(ABSPATH . 'wp-admin/includes/file.php');
		require_once(ABSPATH . 'wp-admin/includes/media.php');
		require_once(ABSPATH . 'wp-admin/includes/image.php');

		// do the validation and storage stuff
		// $post_data can override the items saved to wp_posts table, like post_mime_type, guid, post_parent, post_title, post_content, post_status
		$att_id = media_handle_sideload( $file_array, $post_id, $title, $post_data );

		// If error storing permanently, unlink
		if ( is_wp_error($att_id) ) {
			@unlink($file_array['tmp_name']);   // clean up
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error, %s( %s ), media_handle_sideload returned %s.', $this->getPluginTextDomain() ), __FUNCTION__, $url, $att_id->get_error_message() ) );
		}

		// now set alt tag
		update_post_meta( $att_id, '_wp_attachment_image_alt', $title );

		// set as post thumbnail if desired
		if ($thumb) {
			set_post_thumbnail($post_id, $att_id);
		}

		@unlink($file_array['tmp_name']);   // clean up

		$newFile = wp_get_attachment_url( $att_id );

		return array( 'attachID' => $att_id, 'newFile' => $newFile );
	}

	/* ====================================================================================== */

	/**
	 * find Featured Image for a post. Look for attachment or image in post content
	 * @param $rulesetID
	 * @param $post
	 * return true if successful or false if fail or no image found
	*/
	private function findFeaturedImage( $rulesetID, $post ) {

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s looking for Featured Image for post id %s.', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, $post->ID ) );

		// look for any images in post_content:
		preg_match_all( '/<img.+src=[\'"](?P<src>.+?)[\'"].*>/i', $post->post_content, $matches );
		$imageUrl = null;	// in case no matches at all
		foreach( $matches['src'] as $imageUrl ) {

			$result = $this->fetchImageAsAttachment( $imageUrl, $post->ID, true, $post->post_title );

			if ( ! $result || is_wp_error( $result['attachID'] ) )
				continue;	// try next image found

			// otherwise we successfully downloaded, attached, and set this featured image
			$this->addLogMessage( sprintf( esc_html__( 
				'Ruleset #%s, post id %s featured image set to image url [%s] with attachment id %s.', $this->getPluginTextDomain() ), 
					$rulesetID, 
					"<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", 
					"<a target='_blank' href='" . $result['newFile'] . "'>" . $result['newFile'] . "</a>",
					"<a target='_blank' href='" . get_post_permalink( $result['attachID'] ) . "'>" . $result['attachID'] . "</a>"
				) );
			return true;
		}

		// we ran out of images to check!
		$this->addLogMessage( sprintf( esc_html__( 
				'WAP Warning: Ruleset #%s, post id %s no valid image found attached or in post content to set the Featured Image.', $this->getPluginTextDomain() ), 
					$rulesetID, 
					"<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>"
		) );

		return false;
	}

	/* ====================================================================================== */

	/**
	 * determine box coords for text to insert
	 * @param $fontSize
	 * @param $fontAngle
	 * @param $fontFile
	 * @param $font_text
	 * return array of coords or false on error
	 * taken from https://www.php.net/manual/en/function.imagettfbbox.php#97357
	*/
	private function calculateTextBox( $fontSize, $fontAngle, $fontFileName, $text ) {

		$rect = imagettfbbox( $fontSize, $fontAngle, $fontFileName, $text );
		$minX = min(array($rect[0],$rect[2],$rect[4],$rect[6]));
		$maxX = max(array($rect[0],$rect[2],$rect[4],$rect[6]));
		$minY = min(array($rect[1],$rect[3],$rect[5],$rect[7]));
		$maxY = max(array($rect[1],$rect[3],$rect[5],$rect[7]));
   
		$resultBox = array(
			"left"   => abs($minX) - 1,
			"top"    => abs($minY) - 1,
			"width"  => $maxX - $minX,
			"height" => $maxY - $minY,
			"box"    => $rect
		);
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - TextBox result for [%s] using [%s] and font size %d: %s', $this->getPluginTextDomain() ), $text, $fontFileName, $fontSize, print_r($resultBox,true) ) );

		return $resultBox;
	}

	/* ====================================================================================== */
	/**
	 * create blank canvas image
	 * @param $width
	 * @param $height
	 * return GD image
	*/
	private function createCanvas( $width, $height ) {

		if ( $width && ! $height )
			$height = $width;
		if ( ! $width && $height )
			$width = $height;

		if ( $width && $width < get_option( 'thumbnail_size_w' ) )
			$width = get_option( 'thumbnail_size_w' );
		if ( $height && $height < get_option( 'thumbnail_size_h' ) )
			$height = get_option( 'thumbnail_size_h' );

		$msg = empty( $width) && empty( $height ) ? "Will use existing featured image size, if available." : "";
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, size %dx%d. %s', $this->getPluginTextDomain() ), __FUNCTION__, $width, $height, $msg ) );

		// create blank image canvas if given dimensions
		if ( $width & $height )
			return imagecreatetruecolor( $width, $height );

		// if no dimensions
		return false;
	}

	/* ====================================================================================== */

	/**
	 * set featured image from setting
	 * @param $rulesetID
	 * @param $post
	 * return true if successful or false if fail
	*/
	private function setFeaturedImage( $rulesetID, $post ) {

		$backgroundImages = $this->getRulesetOption( $rulesetID, 'FeaturedImageBGImage', "");

		if ( ! $backgroundImages )	// should never happen
			return false;

		// change string to array and filter out any empty values
		$images = array_filter( explode( ',', $backgroundImages ) );

		// now pick one
		$backgroundImage = array_rand( array_flip( $images ), 1 );

		$result = set_post_thumbnail( $post->ID, $backgroundImage );

		if ( false === $result ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, set_post_thumbnail for post id %d using image %d returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $post->ID, $backgroundImage ) );
		}

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s, %s image id %d set to featured image for post id %d.', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, __FUNCTION__, $backgroundImage, $post->ID ) );

		return true;
	}

	/* ====================================================================================== */

	/**
	 * create featured jpg image from post title. 
	 * @param $rulesetID
	 * @param $post
	 * return true if successful or false if fail
	*/
	private function createImageFromText( $rulesetID, $post ) {

		do_action( $this->getPluginTextDomain() . '-createImageFromText' );

		// check for GD freetype support
		$gd = gd_info();
		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s gd_info() is %s.', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, print_r($gd,true) ) );
		if ( true !== $gd['FreeType Support'] ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, PHP was not compiled with freetype library support for fonts! Cannot create text image using freetype!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__ ) );
		}

		$angle = 0;
		$x = $y = 0;
		$fontDir = $this->getPluginDir() . 'fonts' . DIRECTORY_SEPARATOR;
		$wpUploadData = wp_upload_dir();
		$userDir = $wpUploadData['basedir'] . DIRECTORY_SEPARATOR;
		$imagePath = null;
		$quality = 100;

		// possible params:
		$align = 'center-center';

		$headerText = $this->getRulesetOption( $rulesetID, 'FeaturedImageHeaderText', "");
		$footerText = $this->getRulesetOption( $rulesetID, 'FeaturedImageFooterText', "");
		$width = $this->getRulesetOption( $rulesetID, 'FeaturedImageWidth', get_option( 'thumbnail_size_w' ) );
		$height = $this->getRulesetOption( $rulesetID, 'FeaturedImageHeight', get_option( 'thumbnail_size_h' ) );
		$padding = $this->getRulesetOption( $rulesetID, 'FeaturedImagePadding', 10 );
		$backgroundImage = get_post_thumbnail_id( $post );
		$background = $this->getRulesetOption( $rulesetID, 'FeaturedImageBGColor', "#ffffff");
		$foreground = $this->getRulesetOption( $rulesetID, 'FeaturedImageTextColor', "#000000");
		$fontName = $this->getRulesetOption( $rulesetID, 'FeaturedImageFontType', "OpenSans-Regular.ttf");
		$fontFileName = file_exists( $fontDir . $fontName ) ? $fontDir . $fontName : $userDir . $fontName;
		$fontSize = $this->getRulesetOption( $rulesetID, 'FeaturedImageFontSize', 22 );
		$textTransform = $this->getRulesetOption( $rulesetID, 'FeaturedImageTextTransform', 'none' );
		$textEffect = $this->getRulesetOption( $rulesetID, 'FeaturedImageTextEffect', 'none' );
		$textEncoding = $this->getRulesetOption( $rulesetID, 'FeaturedImageTextEncoding', 'UTF-8' );
		
		$maxLength = $this->getRulesetOption( $rulesetID, 'FeaturedImageMaxTextLength', 255 );
		$lineSpacing = $this->getRulesetOption( $rulesetID, 'FeaturedImageLineSpacing', 1.5 );
		$fileType = $this->getRulesetOption( $rulesetID, 'FeaturedImageFileType', 'png' );

		if( ! preg_match('/^#[a-f0-9]{6}$/i', $background ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, background hex color %s not recognized. Defaulting to #ffffff!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $background ) );
			$background = "#ffffff";
		}
		if( ! preg_match('/^#[a-f0-9]{6}$/i', $foreground ) ) {
			$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, text hex color %s not recognized. Defaulting to #000000!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $foreground ) );
			$foreground = "#000000";
		}

		if ( ! function_exists( 'imagettftext' ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, PHP was not compiled with freetype library support for fonts!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__ ) );
		}

		$image = $this->createCanvas( $width, $height );

		// check for background image to place on canvas
		if ( ! empty( $backgroundImage ) ) {
			$bgFileName = get_attached_file( $backgroundImage );
			$bgFileType = exif_imagetype( $bgFileName );
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s loading image %d, %s to use for background for post id %s.', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, $backgroundImage, $bgFileName, $post->ID ) );
			switch ( $bgFileType ) {
				case IMAGETYPE_GIF:
					$func = 'imagecreatefromgif';
					break;
				case IMAGETYPE_JPEG:
					$func = 'imagecreatefromjpeg';
					break;
				case IMAGETYPE_PNG:
					$func = 'imagecreatefrompng';
					break;
				case IMAGETYPE_BMP:
					$func = 'imagecreatefrombmp';
					break;
				default:
					// type not supported!
					$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, %s filetype is not supported! GIF, JPG, PNG, BMP only.', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, basename( $bgFileName ) ) );
					$backgroundImage = $func = null; // use hex color instead below
					break;
			}

			if ( $func ) {
				$bgImage = $func( $bgFileName );
				if ( false === $bgImage )
					$this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, %s on %s returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $func, $bgFileName ) );
				else {
					$imageInfo = wp_getimagesize( $bgFileName );
					if ( $imageInfo ) {
						if ( ! $width || ! $height ) {
							// no size declared, create canvas with bgImage size 
							$width = $imageInfo[0];
							$height = $imageInfo[1];
							$image = $this->createCanvas( $width, $height );
						}
						// resample bgImage into canvas
						if ( false !== $image ) {
							imagecopyresampled( $image, $bgImage, 0, 0, 0, 0,
								$width, $height, $imageInfo[0], $imageInfo[1] );
						}
					}
				}
			}
		}

		if ( false === $image ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, image GD is false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__ ) );
		}

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s creating Featured Image %dx%d with padding %spx from title for post id %s.', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, $width, $height, $padding, $post->ID ) );

		// create image background color and fill image with it
		if ( empty( $backgroundImage ) && ! empty( $background ) ) {
		list($r, $g, $b) = sscanf( $background, "#%02x%02x%02x" );
		$backgroundColor = imagecolorallocate( $image, $r, $g, $b );
		if ( false === $backgroundColor ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, imagecolorallocate for bc %s returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $background ) );
		}
		if ( false === imagefill( $image, 0, 0, $backgroundColor ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, imagefill returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__ ) );
		}
		}

		// create image foreground color
		list($r, $g, $b) = sscanf( $foreground, "#%02x%02x%02x" );
		$textColor = imagecolorallocate( $image, $r, $g, $b );
		if ( false === $textColor ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, imagecolorallocate for fc %s returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $foreground ) );
		}

		// create shadow color if needed based on text color
		$effectColor = null;
		if ( $textEffect != 'none' ) {
			$HSL = ( max( $r, $g, $b ) + min( $r, $g, $b ) ) / 510.0;
			if ( $HSL >= 0.8 ) 	// light
				$shadow = "#000000";
			else
				$shadow = "#ffffff";
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s - setting shadow color to [%s] because of HSL %s', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, $shadow, $HSL ) );
			list($r, $g, $b) = sscanf( $shadow, "#%02x%02x%02x" );
			$effectColor = imagecolorallocate( $image, $r, $g, $b );
			if ( false === $effectColor ) {
				return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, imagecolorallocate for shadow %s returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $shadow ) );
			}
		}
		
		// get and trim post title
		$text = apply_filters( $this->getPluginTextDomain() . '-create-image-title', mb_convert_encoding( $post->post_title, $textEncoding ) );
		$text = substr( $text, 0, $maxLength );

		// any text transform
		if ( $textTransform == 'uppercase' ) {
			$text = strtoupper( $text );
			$headerText = strtoupper( $headerText );
			$footerText = strtoupper( $footerText );
		}
		else if ( $textTransform == 'lowercase' ) {
			$text = strtolower( $text );
			$headerText = strtolower( $headerText );
			$footerText = strtolower( $footerText );
		}
		else if ( $textTransform == 'ucwords' ) {
			$text = strtolower( $text );
			$headerText = strtolower( $headerText );
			$footerText = strtolower( $footerText );
			$text = ucwords( $text );
			$headerText = ucwords( $headerText );
			$footerText = ucwords( $footerText );
		}

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s - adding text [%s] in color [%s] to image', $this->getPluginTextDomain() ), __FUNCTION__, $rulesetID, $text, $foreground ) );

		// how many chars can fit per line inside the image? then split lines
		while ( $fontSize ) {
			$textBox = $this->calculateTextBox( $fontSize, $angle, $fontFileName, $text );
			$charWidth = $textBox['width'] / strlen( $text );	// in pixels
			$charHeight = $textBox['height'];	// in pixels
			$charsPerLine = floor( ( $width - $padding * 2 ) / ( $charWidth + 1 ) );

			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - charWidth is %d, charHeight is %d, max charsPerLine is %d', $this->getPluginTextDomain() ), $charWidth, $charHeight, $charsPerLine ) );

			$output = wordwrap( $text, $charsPerLine, '\n', false );
			$lines = explode( '\n', $output );
			if ( ! empty( $headerText ) ) 
				array_unshift( $lines, $headerText );
			if ( ! empty( $footerText ) ) 
				array_push( $lines, $footerText );
			$lineCount = count( $lines );

			// will these split lines fit inside image?
			$linesHeight = ceil( ( ( $lineCount-1 )  * $charHeight * $lineSpacing ) + $charHeight );
			if ( $linesHeight <= $height - ( $padding * 2 ) ) {
				break;
			}
			// if it doesn't, reduce font size and try again
			$fontSize--;
			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - total text height %d > image height-padding %d, reducing font size to %d', $this->getPluginTextDomain() ), $linesHeight, $height - ( $padding * 2 ), $fontSize ) );
		}

		if ( ! $fontSize ) {	// should never happen
			return;
		}

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - fontSize is %d, total text height is %d, %d lines are %s', $this->getPluginTextDomain() ), $fontSize, $linesHeight, $lineCount, print_r($lines,true) ) );

		// calculate coords for writing each line into image
		foreach ( $lines as $lineNum => $line ) {
			$textBox = $this->calculateTextBox( $fontSize, $angle, $fontFileName, $line );
			// width should change per line, but height per line should be same
			$thisLineSpacing = $lineNum ? $lineSpacing : 1;

			switch( $align ) {
				case 'center-center':
//JPH more options?
					$x = ( ( $width - $textBox['width'] ) / 2 ) - $textBox['left'];
					$y = ( ( $height - $linesHeight ) / 2 ) + $charHeight + ( $charHeight * $lineNum * $lineSpacing );
					break;
			}

			if ( $x < 0 )		// double verify this never happens
				$x = 0;
			if ( $y < $charHeight )	// double verify this never happens
				$y = $charHeight;

			$this->addLogMessage( sprintf( esc_html__( 'DEBUG - writing line [%s] to coords %d, %d', $this->getPluginTextDomain() ), $line, $x, $y ) );

			// create the shadow text in the image, if used
			if ( $textEffect == 'drop-shadow' )
				imagettftext( $image, $fontSize, $angle, $x+2, $y+2, $effectColor, $fontFileName, $line );
			else if ( $textEffect == 'stroke' ) {
				// base on http://www.johnciacia.com/2010/01/04/using-php-and-gd-to-add-border-to-text/
				$px = 2;
				for($c1 = ($x-abs($px)); $c1 <= ($x+abs($px)); $c1++)
					for($c2 = ($y-abs($px)); $c2 <= ($y+abs($px)); $c2++)
						imagettftext( $image, $fontSize, $angle, $c1, $c2, $effectColor, $fontFileName, $line );
			}

			// create the foreground text in the image
			if ( ! $box = imagettftext( $image, $fontSize, $angle, $x, $y, $textColor, $fontFileName, $line ) ) {
				return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, imagettftext at coords %d, %d with font [%s] returned false! Check server logs for error.', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $x, $y, $fontFileName ) );
			}
		} // end foreach

		// calculate unique filename for image
		$wpUploadData = wp_upload_dir();
		$wpUploadPath = $wpUploadData['path'];
		$slug = wp_unique_post_slug( sanitize_title( $text ), $post->ID, $post->post_status, $post->post_type, $post->post_parent );
		$filename = strtolower( wp_unique_filename( $wpUploadPath, $text . "." . $fileType ) );
		$imagePath = $wpUploadPath . "/" . $filename;
		$imageUrl = $wpUploadData['url'] . '/' . $filename;

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - function %s, Ruleset #%s - saving image to [%s]', $this->getPluginTextDomain() ), 
				__FUNCTION__, 
				$rulesetID, 
				"<a target='_blank' href='" . $imageUrl . "'>" . $imageUrl . "</a>"
		) );

		// save the image!
		if ( $fileType == "png" ) {
			if ( false === imagepng( $image, $imagePath ) ) {
				return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, imagepng to path [%s] returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $imagePath ) );
			}
			$mimeType = 'image/png';
		}
		else if ( $fileType == "jpg" ) {
			if ( false === imagejpeg( $image, $imagePath, $quality ) ) {
				return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, imagejpeg to path [%s] returned false!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $imagePath ) );
			}
			$mimeType = 'image/jpeg';
		}

		if ( ! file_exists( $imagePath ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, image %s does not exist!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $imagePath ) );
		}

		// attach image to post as featured
		// Prepare an array of post data for the attachment.
		$attachment = array(
			'guid'			=> $wpUploadData['url'] . '/' . basename( $filename ), 
			'file'			=> $imagePath,
			'post_mime_type'	=> $mimeType,
			'post_parent'		=> $post->ID,
			'post_title'		=> $post->post_title,
			'post_name'		=> $post->post_title
		);

		// Insert the attachment.
		$attachID = wp_insert_attachment( $attachment, $imagePath, $post->ID, true );

		if ( is_wp_error( $attachID ) ) {
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error: Ruleset #%s - function %s, wp_insert_attachment for [%s] returned %s!', $this->getPluginTextDomain() ), $rulesetID, __FUNCTION__, $imagePath, $attachID->get_error_message() ) );
		}

		// Make sure that this file is included, as wp_generate_attachment_metadata() depends on it.
		require_once( ABSPATH . 'wp-admin/includes/image.php' );

		// Generate the metadata for the attachment, and update the database record.
		$attach_data = wp_generate_attachment_metadata( $attachID, $imagePath );
		wp_update_attachment_metadata( $attachID, $attach_data );

		// now set alt tag
		update_post_meta( $attachID, '_wp_attachment_image_alt', $post->post_title );

		set_post_thumbnail( $post->ID, $attachID );

		$this->addLogMessage( sprintf( esc_html__( 
			'Ruleset #%s, post id %s Created featured image set to attachment id %s.', $this->getPluginTextDomain() ), 
				$rulesetID, 
				"<a target='_blank' href='" . get_post_permalink( $post->ID ) . "'>" . $post->ID . "</a>", 
				"<a target='_blank' href='" . get_post_permalink( $attachID ) . "'>" . $attachID . "</a>"
		) );

		return true;

	}

	/* ====================================================================================== */

	/**
	 * update image preview when selected in media library modal window 
	*/
	public function getImage() {
		$image = '';
		if ( isset( $_GET['id'] ) ) {

			$gets = filter_input( INPUT_GET, 'id' );
			$ids = explode( ',', filter_input( INPUT_GET, 'id' ) );
			foreach ( $ids as $id ) {
				if ( empty( $id ) ) continue;
				$image .= wp_get_attachment_image( $id, 'thumbnail', false, array( 'id' => 'img-responsive' ) );
			}

			$data = array(
				'image'		=> "<p id='previewImage'>" . $image . "</p>"
			);
			wp_send_json_success( $data );
		}
		else	
			wp_send_json_error();
	}

	/* ====================================================================================== */


	/**
	 * make and send query to tweetpik API  
	 * @param $tweetID
	 * return nothing
	 */
	private function tweetPik( $tweetID ) {
		$apiKey = '5b233935-88a5-4a14-9f1d-622518a393e3';
		$themeID = null;

$tweetID = '1370397279279054848';

		$url = 'https://tweetpik.com/api/images';

		$headers = array(
			'Content-Type'	=> 'application/json',
			'Authorization' => $apiKey
		);

		$data = array(
			'tweetId'	=> $tweetID,
			'themeId'	=> $themeID
		);

		$args = array(
			'headers'	=> $headers,
			'body'		=> json_encode( $data ),
			'method'	=> 'POST'
		);

		$response = wp_remote_post( $url, $args );

error_log("JPH request is " . print_r($args,true));

		if ( is_wp_error( $response ) )
			return $this->addLogMessage( sprintf( esc_html__( 'WAP Error - tweetpik returned %s.', $this->getPluginTextDomain() ), $response->get_error_message() ) );

		$this->addLogMessage( sprintf( esc_html__( 'DEBUG - %s request [%s] returned [%s].', $this->getPluginTextDomain() ), __FUNCTION__, print_r($args,true), print_r($response,true) ) );
error_log("JPH tweetpik reaponse: " . print_r($response,true));
	}

	/* ====================================================================================== */

} // end class WpAutoPoster_Plugin extends WpAutoPoster_LifeCycle 
